from __future__ import annotations

from configparser import ConfigParser

import pytest

from pyaerocom.io.ebas_file_index import EbasSQLRequest
from pyaerocom.io.ebas_varinfo import EbasVarInfo

TESTDATA = [
    ("DEFAULT", None, None, None, None, None, 1),
    (
        "sc550aer",
        ["aerosol_light_scattering_coefficient"],
        ["aerosol", "pm10"],
        None,
        None,
        None,
        1.0,
    ),
    (
        "sc440aer",
        ["aerosol_light_scattering_coefficient"],
        ["aerosol", "pm10"],
        None,
        None,
        None,
        1.0,
    ),
    (
        "sc700aer",
        ["aerosol_light_scattering_coefficient"],
        ["aerosol", "pm10"],
        None,
        None,
        None,
        1.0,
    ),
    ("sc550dryaer", None, None, None, None, ["sc550aer", "scrh"], 1),
    ("sc440dryaer", None, None, None, None, ["sc440aer", "scrh"], 1),
    ("sc700dryaer", None, None, None, None, ["sc700aer", "scrh"], 1),
    (
        "sc550lt1aer",
        ["aerosol_light_scattering_coefficient"],
        ["pm25", "pm1"],
        None,
        None,
        None,
        1.0,
    ),
    (
        "bsc550aer",
        ["aerosol_light_backscattering_coefficient"],
        ["aerosol", "pm10", "pm25"],
        None,
        None,
        None,
        1.0,
    ),
    (
        "ac550aer",
        ["aerosol_absorption_coefficient"],
        ["aerosol", "pm10"],
        ["filter_absorption_photometer"],
        None,
        None,
        1.0,
    ),
    (
        "ac550dryaer",
        None,
        None,
        ["filter_absorption_photometer"],
        None,
        ["ac550aer", "acrh"],
        1,
    ),
    (
        "ac550lt1aer",
        ["aerosol_absorption_coefficient"],
        ["pm25", "pm1"],
        ["filter_absorption_photometer"],
        None,
        None,
        1.0,
    ),
    (
        "bsc550dryaer",
        ["aerosol_light_backscattering_coefficient"],
        ["pm10", "pm25", "pm1", "aerosol"],
        ["nephelometer"],
        None,
        None,
        1.0,
    ),
    (
        "scrh",
        ["relative_humidity"],
        ["instrument", "aerosol", "met", "pm10", "pm25", "pm1"],
        None,
        None,
        None,
        1,
    ),
    (
        "acrh",
        ["relative_humidity"],
        ["instrument", "aerosol", "met", "pm10", "pm25", "pm1"],
        None,
        None,
        None,
        1,
    ),
    (
        "concso4",
        ["sulphate_corrected", "sulphate_total"],
        ["aerosol", "pm10", "pm25"],
        None,
        None,
        None,
        1,
    ),
    ("concso2", ["sulphur_dioxide"], ["air"], None, None, None, 1.0),
    ("concpm10", ["pm10_mass"], ["pm10"], None, None, None, 1.0),
    ("concpm25", ["pm25_mass"], ["pm25"], None, None, None, 1.0),
    (
        "concso4t",
        ["sulphate_total"],
        ["aerosol", "pm10", "pm25"],
        None,
        None,
        None,
        1.0,
    ),
    (
        "concso4c",
        ["sulphate_corrected"],
        ["aerosol", "pm10", "pm25"],
        None,
        None,
        None,
        1.0,
    ),
    (
        "concbc",
        ["elemental_carbon"],
        ["pm25", "pm10", "pm1", "aerosol"],
        [
            "denuder",
            "ecoc_monitor",
            "filter_1pack",
            "filter_2pack",
            "high_vol_sampler",
            "impactor",
            "low_vol_sampler",
            "lvs_denuder_single",
            "lvs_denuder_tandem",
            "lvs_QBQ",
            "lvs_single",
            "lvs_single_twin",
            "lvs_teflon",
        ],
        ["arithmetic mean", "median"],
        None,
        1.0,
    ),
    (
        "conceqbc",
        ["equivalent_black_carbon"],
        ["aerosol", "pm1", "pm10", "pm25"],
        ["filter_absorption_photometer"],
        None,
        None,
        1,
    ),
    (
        "conctc",
        ["total_carbon"],
        ["pm25", "pm10", "aerosol"],
        None,
        ["arithmetic mean", "median"],
        None,
        1.0,
    ),
    (
        "concoa",
        ["organic_carbon"],
        ["pm25", "pm10", "aerosol", "pm1"],
        None,
        ["arithmetic mean", "median"],
        None,
        1.4,
    ),
    (
        "concoc",
        ["organic_carbon"],
        ["pm25", "pm10", "aerosol", "pm1"],
        None,
        ["arithmetic mean", "median"],
        None,
        1,
    ),
    (
        "concss",
        ["sodium"],
        ["pm10", "aerosol", "pm25", "pm1", "air"],
        None,
        None,
        None,
        3.27,
    ),
    ("concnh3", ["ammonia"], ["air"], None, None, None, 1.0),
    ("concno3", ["nitrate"], ["pm10", "aerosol", "pm25"], None, None, None, 1.0),
    ("concnh4", ["ammonium"], ["pm10", "aerosol", "pm25"], None, None, None, 1.0),
    ("concNhno3", ["nitric_acid"], ["air"], None, None, None, 1.0),
    (
        "concNtno3",
        ["sum_nitric_acid_and_nitrate"],
        ["air+aerosol"],
        None,
        None,
        None,
        1.0,
    ),
    ("concno2", ["nitrogen_dioxide"], ["air"], None, None, None, 1.0),
    ("conco3", ["ozone"], ["air"], None, None, None, 1),
    ("concco", ["carbon_monoxide"], ["air"], None, None, None, 1.0),
    (
        "concprcpoxs",
        ["sulphate_corrected", "sulphate_total"],
        ["precip"],
        None,
        None,
        None,
        1.0,
    ),
    ("concprcpoxn", ["nitrate"], ["precip"], None, None, None, 1.0),
    ("concprcprdn", ["ammonium"], ["precip"], None, None, None, 1.0),
    ("wetoxs", None, None, None, None, ["concprcpoxs"], 1),
    ("wetoxn", None, None, None, None, ["concprcpoxn"], 1),
    ("wetrdn", None, None, None, None, ["concprcprdn"], 1),
    ("wetoxsc", None, None, None, None, ["concprcpoxsc"], 1),
    ("wetoxst", None, None, None, None, ["concprcpoxst"], 1),
    ("wetna", None, None, None, None, ["concprcpna"], 1),
    ("proxydryoxs", None, None, None, None, ["concprcpoxs"], 1),
    ("proxydryoxn", None, None, None, None, ["concprcpoxn"], 1),
    ("proxydryrdn", None, None, None, None, ["concprcprdn"], 1),
    ("proxydryo3", None, None, None, None, ["vmro3"], 1),
    ("wetno3", None, None, None, None, ["concprcpno3"], 1),
    ("wetnh4", None, None, None, None, ["concprcpnh4"], 1),
    ("wetso4", None, None, None, None, ["concprcpso4"], 1),
    (
        "prmm",
        ["precipitation_amount_off", "precipitation_amount"],
        ["precip"],
        None,
        None,
        None,
        1.0,
    ),
]


def test_init_empty():
    with pytest.raises(TypeError) as e:
        EbasVarInfo()
    assert "missing 1 required positional argument" in str(e.value)


def test_open_config():
    assert isinstance(EbasVarInfo.open_config(), ConfigParser)


def test_var_name_aerocom():
    assert EbasVarInfo("sconcno3").var_name_aerocom == "concno3"


@pytest.fixture
def info(var_name: str | None) -> EbasVarInfo:
    if var_name is None:
        info = EbasVarInfo("concpm10")
        info.component = None
        return info
    return EbasVarInfo(var_name)


@pytest.mark.parametrize(
    "var_name,component,matrix,instrument,statistics,requires,scale_factor", TESTDATA
)
def test_varinfo(
    info: EbasVarInfo, component, matrix, instrument, statistics, requires, scale_factor
):
    assert info.component == component
    assert info.matrix == matrix
    assert info.instrument == instrument
    print(info.statistics, statistics, info.statistics == statistics)
    assert info.statistics == statistics
    assert info.requires == requires
    assert info.scale_factor == scale_factor


def test_to_dict():
    info = EbasVarInfo("concpm10")
    dic = info.to_dict()
    assert isinstance(dic, dict)
    for key, val in dic.items():
        assert getattr(info, key) == val


@pytest.mark.parametrize(
    "var_name,constraints",
    [
        ("concpm10", {}),
        ("concpm10", {"bla": 42}),
        # ToDo: the following example should actually be checked and maybe raise an Exception already here
        # (i.e. start_date is a valid constraint but 42 is not allowed as input)
        ("concpm10", {"start_date": 42}),
    ],
)
def test_make_sql_request(info: EbasVarInfo, constraints: dict):
    request = info.make_sql_request(**constraints)
    assert isinstance(request, EbasSQLRequest)


@pytest.mark.parametrize(
    "var_name,exception,error",
    [
        (
            None,
            AttributeError,
            "At least one component (Ebas variable name) must be specified",
        ),
        (
            "sc700dryaer",
            ValueError,
            "This variable sc700dryaer requires other variables for reading",
        ),
        (
            "sc550dryaer",
            ValueError,
            "This variable sc550dryaer requires other variables for reading",
        ),
    ],
)
def test_make_sql_request_error(info: EbasVarInfo, exception: type[Exception], error: str):
    with pytest.raises(exception) as e:
        info.make_sql_request()
    assert str(e.value).startswith(error)


@pytest.mark.parametrize(
    "var_name,num",
    [
        ("concpm10", 1),
        ("sc700dryaer", 2),
        ("sc550dryaer", 2),
        ("sc440dryaer", 2),
    ],
)
def test_make_sql_requests(info: EbasVarInfo, num: int):
    requests = info.make_sql_requests()
    assert isinstance(requests, dict)
    assert len(requests) == num
    assert all(isinstance(req, EbasSQLRequest) for req in requests.values())


def test___str__():
    var_info_str = str(EbasVarInfo("concpm10"))
    assert var_info_str.startswith("\nPyaerocom EbasVarInfo")
    assert "var_name: concpm10" in var_info_str



from __future__ import annotations

from datetime import datetime

import pytest

from pyaerocom.scripts.cams2_83.evaluation import EvalType, date_range


@pytest.mark.parametrize(
    "eval_type,start_date,end_date,result",
    [
        pytest.param(
            "long",
            datetime(2021, 12, 1),
            datetime(2024, 5, 31),
            ["2021-2024", "2021", "2022", "2023", "2024"],
            id="long",
        ),
        pytest.param(
            "long",
            datetime(2024, 3, 1),
            datetime(2024, 8, 31),
            ["20240301-20240831"],
            id="long",
        ),
        pytest.param(
            "week",
            datetime(2023, 12, 28),
            datetime(2024, 1, 12),
            ["20231228-20240112"],
            id="week",
        ),
        pytest.param("day", datetime(2024, 1, 12), datetime(2024, 1, 12), ["20240112"], id="day"),
    ],
)
def test_periods(eval_type: str, start_date: datetime, end_date: datetime, result: tuple):
    eval = EvalType(eval_type)
    assert eval.periods(start_date, end_date) == result


@pytest.mark.parametrize(
    "eval_type,start_date,end_date,error",
    [
        pytest.param(
            "season",
            datetime(2023, 12, 28),
            datetime(2023, 12, 12),
            "End date should be ⩾ start_date",
            id="invalid",
        ),
        pytest.param(
            "week",
            datetime(2023, 12, 28),
            datetime(2024, 1, 1),
            "Evaluation type 'week' should have",
            id="week",
        ),
        pytest.param(
            "day",
            datetime(2024, 1, 12),
            datetime(2024, 2, 12),
            "Evaluation type 'day' should have the same",
            id="day",
        ),
    ],
)
def test_check_dates(eval_type: str, start_date: datetime, end_date: datetime, error: str):
    eval = EvalType(eval_type)
    with pytest.raises(ValueError) as excinfo:
        eval.check_dates(start_date, end_date)
    assert error in str(excinfo.value)


@pytest.mark.parametrize(
    "start_date,end_date,result",
    [
        pytest.param(
            datetime(2023, 12, 28),
            datetime(2023, 12, 28),
            (datetime(2023, 12, 28, 0, 0),),
            id="1d",
        ),
        pytest.param(
            datetime(2023, 12, 28),
            datetime(2023, 12, 30),
            (
                datetime(2023, 12, 28, 0, 0),
                datetime(2023, 12, 29, 0, 0),
                datetime(2023, 12, 30, 0, 0),
            ),
            id="3d",
        ),
    ],
)
def test_date_range(start_date: datetime, end_date: datetime, result: tuple[datetime, ...]):
    assert date_range(start_date=start_date, end_date=end_date) == result



from pyaerocom.config import ALL_REGION_NAME

MODELS = {"TM5-AP3-CTRL": dict(model_id="TM5-met2010_CTRL-TEST")}

OBS_GROUNDBASED = {
    "AERONET-Sun": dict(
        obs_id="AeronetSunV3L2Subset.daily",
        obs_vars=("od550aer",),
        obs_vert_type="Column",
    )
}

CFG = dict(
    model_cfg=MODELS,
    obs_cfg=OBS_GROUNDBASED,
    json_basedir="PATH_TO_AEROVAL_OUT/data",
    coldata_basedir="PATH_TO_AEROVAL_OUT/coldata",
    # if True, existing colocated data files will be deleted
    reanalyse_existing=True,
    raise_exceptions=True,
    only_json=False,
    add_model_maps=True,
    only_model_maps=False,
    clear_existing_json=False,
    # Regional filter for analysis
    filter_name=f"{ALL_REGION_NAME}-wMOUNTAINS",
    # colocation frequency (no statistics in higher resolution can be computed)
    ts_type="daily",
    map_zoom="World",
    freqs=["daily", "monthly", "yearly"],
    periods=["2010"],
    main_freq="monthly",
    zeros_to_nan=False,
    min_num_obs=None,
    colocate_time=False,
    obs_remove_outliers=False,
    add_seasons=False,
    model_remove_outliers=False,
    harmonise_units=True,
    regions_how="default",  #'default',#'country',
    annual_stats_constrained=True,
    proj_id="test",
    exp_id="exp1",
    exp_name="AeroVal test experiment 1",
    exp_descr=("Very simple setup to test basic stuff in AeroVal"),
    exp_pi="Jonas Gliss",
    public=True,
    # directory where colocated data files are supposed to be stored
    weighted_stats=True,
)



import pytest

from pyaerocom.metastandards import AerocomDataID, DataSource, StationMetaData


def test_datasource_empty():
    ds = DataSource()

    keys_sorted = [
        "data_id",
        "data_level",
        "data_product",
        "data_version",
        "dataset_name",
        "framework",
        "instr_vert_loc",
        "revision_date",
        "stat_merge_pref_attr",
        "ts_type_src",
        "website",
    ]
    assert sorted(list(ds)) == keys_sorted
    assert list(set(ds.values())) == [None]


@pytest.mark.parametrize(
    "data_id,dataset_name,data_product,data_version,data_level,revision_date,stat_merge_pref_attr",
    [
        ("AeronetSunV3Lev2.daily", "AERONET", "Sun", 3.0, 2.0, None, None),
        ("EBASMC", "EBAS", None, None, None, None, "revision_date"),
        ("EARLINET", "EARLINET", None, None, None, None, None),
    ],
)
def test_datasource(
    data_id,
    dataset_name,
    data_product,
    data_version,
    data_level,
    revision_date,
    stat_merge_pref_attr,
):
    ds = DataSource(data_id=data_id)
    assert ds["dataset_name"] == dataset_name
    assert ds["data_product"] == data_product
    assert ds["data_version"] == data_version
    assert ds["data_level"] == data_level
    assert ds["revision_date"] == revision_date
    assert ds["stat_merge_pref_attr"] == stat_merge_pref_attr


def test_stationmetadata():
    meta = StationMetaData()

    assert isinstance(meta, DataSource)
    assert sorted(meta.keys(), key=str.casefold) == [
        "altitude",
        "country",
        "country_code",
        "data_id",
        "data_level",
        "data_product",
        "data_version",
        "dataset_name",
        "filename",
        "framework",
        "instr_vert_loc",
        "instrument_name",
        "latitude",
        "longitude",
        "PI",
        "revision_date",
        "stat_merge_pref_attr",
        "station_id",
        "station_name",
        "ts_type",
        "ts_type_src",
        "website",
    ]


@pytest.mark.parametrize(
    "data_id,values,test_addstuff",
    [
        ("NorESM2-met2010_AP3-CTRL", ["NorESM2", "met2010", "AP3", "CTRL"], True),
        ("Blaaa", ["Blaaa", "", "", ""], False),
        ("Bla-blub2010_blablub-bla", ["Bla-blub2010", "", "blablub", "bla"], False),
        ("Bla-met2042_blablub-bla", ["Bla", "met2042", "blablub", "bla"], False),
    ],
)
def test_aerocomdataid(data_id, values, test_addstuff):
    data_id = AerocomDataID(data_id)

    assert data_id.values == values

    if test_addstuff:
        dd = data_id.to_dict()

        assert dd["model_name"] == values[0]
        assert dd["meteo"] == values[1]
        assert dd["experiment"] == values[2]
        assert dd["perturbation"] == values[3]

        data_id1 = AerocomDataID(**dd)

        assert data_id1 == str(data_id)
        assert data_id1 == data_id
        assert data_id1 == "NorESM2-met2010_AP3-CTRL"

        assert AerocomDataID(**dd) == AerocomDataID.from_dict(dd)



import logging
import os
import sys

import numpy as np
import xarray

from pyaerocom import const
from pyaerocom.units import convert_unit
from pyaerocom.exceptions import DataUnitError, DataDimensionError, EprofileFileError
from pyaerocom.io.readungriddedbase import ReadUngriddedBase
from pyaerocom.stationdata import StationData
from pyaerocom.ungriddeddata import UngriddedData
from pyaerocom.units.units_helpers import get_unit_conversion_fac
from pyaerocom.variable import Variable
from pyaerocom.vertical_profile import VerticalProfile
from pathlib import Path

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override


logger = logging.getLogger(__name__)


class ReadEprofile(ReadUngriddedBase):
    """Interface for reading of EARLINET data"""

    #: Mask for identifying datafiles
    _FILEMASK = "*.nc*"

    #: version log of this class (for caching)
    __version__ = "0.01_" + ReadUngriddedBase.__baseversion__

    #: Name of dataset (OBS_ID)
    DATA_ID = const.EPROFILE_NAME

    #: List of all datasets supported by this interface
    SUPPORTED_DATASETS = [const.EPROFILE_NAME]

    #: default variables for read method
    DEFAULT_VARS = ["ec1064aer", "bsc1064aer"]

    CLOUD_FILTERS = {}

    #: all data values that exceed this number will be set to NaN on read. This
    #: is because iris, xarray, etc. assign a FILL VALUE of the order of e36
    #: to missing data in the netcdf files
    _MAX_VAL_NAN = 1e6

    #: variable name of altitude in files
    ALTITUDE_ID = "altitude"

    #: temporal resolution
    # TODO: check this
    TS_TYPE = "hourly"

    #: dictionary specifying the file column names (values) for each Aerocom
    #: variable (keys)
    # TODO: add aod
    VAR_NAMES_FILE = {
        "ec1064aer": "extinction",
        "bsc1064aer": "attenuated_backscatter_0",
    }

    META_NAMES_FILE = dict(
        station_longitude="station_longitude",
        station_latitude="station_latitude",
        station_altitude="station_altitude",
        instrument_name="instrument_id",
        instrument_type="instrument_type",
        comment="comment",
        PI="principal_investigator",
        institution="institution",
        dataset_name="title",
        website="references",
        wavelength_emis="l0_wavelength",
        altitude="altitude",
        history="history",
        overlap_is_corrected="overlap_is_corrected",
    )
    #: metadata keys that are needed for reading (must be values in
    #: :attr:`META_NAMES_FILE`)
    META_NEEDED = [
        "station_longitude_t0",
        "station_latitude_t0",
        "station_altitude_t0",
    ]

    #: Metadata keys from :attr:`META_NAMES_FILE` that are additional to
    #: standard keys defined in :class:`StationMetaData` and that are supposed
    #: to be inserted into :class:`UngriddedData` object created in :func:`read`
    KEEP_ADD_META = [
        "comment",
    ]

    #: Attribute access names for unit reading of variable data
    VAR_UNIT_NAMES = dict(
        extinction=["unit"],  # TODO: needs checking
        attenuated_backscatter_0=["units"],
        altitude=["units"],
    )
    #: Variable names of uncertainty data
    ERR_VARNAMES = dict()

    PROVIDES_VARIABLES = list(DEFAULT_VARS)

    EXCLUDE_CASES = []

    def __init__(self, data_id=None, data_dir=None):
        # initiate base class
        super().__init__(data_id=data_id, data_dir=data_dir)

        #: private dictionary containing loaded Variable instances,
        self._var_info = {}

        #: files that are supposed to be excluded from reading
        self.exclude_files = []

        #: files that were actually excluded from reading
        self.excluded_files = []

        self.is_vertical_profile = True

    @override
    def read_file(self, filename, vars_to_retrieve=None, remove_outliers=True) -> StationData:
        """Read EARLINET file and return it as instance of :class:`StationData`

        Parameters
        ----------
        filename : str
            absolute path to filename to read
        vars_to_retrieve : :obj:`list`, optional
            list of str with variable names to read. If None, use
            :attr:`DEFAULT_VARS`
        remove_outliers : bool
            if True, outliers are removed for each variable using the
            `minimum` and `maximum` attributes for that variable (accessed
            via pyaerocom.const.VARS[var_name]).

        Returns
        -------
        StationData
            dict-like object containing results
        """

        if isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]
        _vars = []

        for var in vars_to_retrieve:
            if var in self.VAR_NAMES_FILE:  # make sure to only read what is supported by this file
                _vars.append(var)
            elif var in self.AUX_REQUIRES:
                _vars.append(var)
            else:
                raise ValueError(f"{var} is not supported")
        vars_to_read, vars_to_compute = self.check_vars_to_retrieve(_vars)

        # create empty data object (is dictionary with extended functionality)
        data_out = StationData()
        data_out["data_id"] = self.data_id
        data_out["ts_type"] = self.TS_TYPE

        # create empty arrays for all variables that are supposed to be read
        # from file
        for var in vars_to_read:
            if var not in self._var_info:
                self._var_info[var] = Variable(var)
        var_info = self._var_info

        # Iterate over the lines of the file
        logger.debug(f"Reading file {filename}")

        with xarray.open_dataset(filename, engine="netcdf4", decode_timedelta=True) as data_in:
            data_out["station_coords"]["longitude"] = data_out["longitude"] = (
                data_in.station_longitude
            )
            data_out["station_coords"]["latitude"] = data_out["latitude"] = (
                data_in.station_latitude
            )
            data_out["altitude"] = (
                data_in.station_altitude_t0 + data_in.altitude.values
            )  # Note altitude is an array for the data, station altitude is different. Moreover, EPROFILE as of 21.05.2025 gives altitude in altitude above ground level, so add the station altitude to get the altitude above sea level
            data_out["station_coords"]["altitude"] = data_in.station_altitude
            data_out["altitude_attrs"] = (
                data_in.altitude.attrs
            )  # get attrs for altitude units + extra

            # get intersection of metadaa in ddataa_out and data_in
            for k, v in self.META_NAMES_FILE.items():
                if v in self.META_NEEDED:
                    _meta = data_in.attrs[v]
                else:
                    try:
                        _meta = data_in.attrs[v]
                    except Exception:  # pragma: no cover
                        _meta = None
                data_out[k] = _meta

            # get metadata expected in StationData but not in data_in's metadata
            data_out["wavelength_emis"] = data_in.l0_wavelength
            data_out["zenith_angle"] = data_in.z_ref.values
            data_out["filename"] = filename

            loc_split = data_in.attrs["site_location"].split(", ")
            data_out["station_name"] = loc_split[0]
            if len(loc_split) > 1:
                data_out["country"] = loc_split[1]

            data_out["dtime"] = data_in.time.values

            for var in vars_to_read:
                data_out["var_info"][var] = {}
                unit_ok = False
                outliers_removed = False
                has_altitude = False

                netcdf_var_name = self.VAR_NAMES_FILE[var]
                # check if the desired variable is in the file
                if netcdf_var_name not in data_in.variables:
                    logger.warning(f"Variable {var} not found in file {filename}")
                    continue

                info = var_info[var]
                arr = data_in.variables[netcdf_var_name]
                val = np.squeeze(np.float64(arr))  # squeeze to 1D array

                # CONVERT UNIT
                unit = None

                unames = self.VAR_UNIT_NAMES[netcdf_var_name]
                for u in unames:
                    if u in arr.attrs:
                        unit = arr.attrs[u]
                if unit is None:
                    raise DataUnitError(f"Unit of {var} could not be accessed in file {filename}")
                if len(unit) > 0:
                    try:
                        to_unit = self._var_info[var].units
                        val = convert_unit(val, from_unit=unit, to_unit=to_unit)
                        unit = to_unit
                        unit_ok = True
                    except Exception as e:
                        logger.warning(
                            f"Failed to convert unit of {var} in file {filename} (EPROFILE): "
                            f"Error: {repr(e)}"
                        )
                else:
                    logger.warning(
                        f"Failed to convert unit of {var} in file {filename} (EPROFILE): Meaningful unit not found in file, so assuming unit_ok = True and that the units in the data are the same as in variables.ini"
                    )
                    unit_ok = True
                    unit = self._var_info[var].units

                if not val.ndim == 2:
                    raise DataDimensionError("EPROFILE data must be two dimensional")
                elif len(val) == 0:
                    continue  # no data
                # Remove NaN equivalent values
                val[val > self._MAX_VAL_NAN] = np.nan

                wvlg = var_info[var].wavelength_nm
                wvlg_str = self.META_NAMES_FILE["wavelength_emis"]

                if not wvlg == data_in.attrs[wvlg_str]:
                    logger.info("No wavelength match")
                    continue

                alt_data = data_in.variables[self.ALTITUDE_ID]

                alt_unit = alt_data.attrs["units"]
                to_alt_unit = const.VARS["alt"].units
                if not alt_unit == to_alt_unit:
                    try:
                        alt_unit_fac = get_unit_conversion_fac(alt_unit, to_alt_unit)
                        alt_data *= alt_unit_fac
                        alt_unit = to_alt_unit
                    except Exception as e:
                        logger.warning(f"Failed to convert unit: {repr(e)}")
                has_altitude = True

                # remove outliers from data, if applicable
                if remove_outliers and unit_ok:
                    # REMOVE OUTLIERS
                    outlier_mask = np.logical_or(val < info.minimum, val > info.maximum)
                    val[outlier_mask] = np.nan
                    outliers_removed = True

                # create instance of ProfileData
                profile = VerticalProfile(
                    data=val,
                    altitude=alt_data.values,
                    dtime=data_in.time.values,
                    var_name=var,
                    data_err=np.nan,  # EPROFILE does not provide error data
                    var_unit=unit,
                    altitude_unit=alt_unit,
                )

                # Write everything into profile
                data_out[var] = profile

            data_out["var_info"][var].update(
                unit_ok=unit_ok,
                err_read=False,  # EPROFILE foes not provide error data
                outliers_removed=outliers_removed,
                has_altitude=has_altitude,
            )
        return data_out

    @override
    def read(
        self,
        vars_to_retrieve=None,
        files=None,
        first_file=None,
        last_file=None,
        remove_outliers=True,
        pattern=None,
    ) -> UngriddedData:
        """Method that reads list of files as instance of :class:`UngriddedData`

        Parameters
        ----------
        vars_to_retrieve : :obj:`list` or similar, optional,
            list containing variable IDs that are supposed to be read. If None,
            all variables in :attr:`PROVIDES_VARIABLES` are loaded
        files : :obj:`list`, optional
            list of files to be read. If None, then the file list is used that
            is returned on :func:`get_file_list`.
        first_file : :obj:`int`, optional
            index of first file in file list to read. If None, the very first
            file in the list is used
        last_file : :obj:`int`, optional
            index of last file in list to read. If None, the very last file
            in the list is used
         pattern : str, optional
            string pattern for file search (cf :func:`get_file_list`)

        Returns
        -------
        UngriddedData
            data object
        """

        if vars_to_retrieve is None:
            vars_to_retrieve = self.DEFAULT_VARS
        elif isinstance(vars_to_retrieve, str):
            vars_to_retrieve = [vars_to_retrieve]

        # if read_err is None:
        #     read_err = self.READ_ERR

        if files is None:
            if len(self.files) == 0:
                self.get_file_list(vars_to_retrieve, pattern=pattern)
            files = self.files

        # turn files into a list because I suspect there may be a bug if you don't do this
        if isinstance(files, str):
            files = [files]

        if first_file is None:
            first_file = 0
        if last_file is None:
            last_file = len(files)

        files = files[first_file : last_file + 1]

        self.read_failed = []

        data_obj = UngriddedData()
        data_obj.is_vertical_profile = True
        col_idx = data_obj.index
        meta_key = -1.0
        idx = 0

        # assign metadata object
        metadata = data_obj.metadata
        meta_idx = data_obj.meta_idx

        # last_station_id = ''
        num_files = len(files)

        disp_each = int(num_files * 0.1)
        if disp_each < 1:
            disp_each = 1

        VAR_IDX = -1
        for i, _file in enumerate(files):
            if i % disp_each == 0:
                logger.info(f"Reading file {i + 1} of {num_files} ({type(self).__name__})")
            try:
                stat = self.read_file(
                    _file,
                    vars_to_retrieve=vars_to_retrieve,
                    remove_outliers=remove_outliers,
                )
                if not any([var in stat.vars_available for var in vars_to_retrieve]):
                    logger.info(
                        f"Station {stat.station_name} contains none of the desired variables. Skipping station..."
                    )
                    continue
                meta_key += 1
                # Fill the metadata dict
                # the location in the data set is time step dependant!
                # use the lat location here since we have to choose one location
                # in the time series plot
                metadata[meta_key] = {}
                metadata[meta_key].update(stat.get_meta())
                for add_meta in self.KEEP_ADD_META:
                    if add_meta in stat:
                        metadata[meta_key][add_meta] = stat[add_meta]
                # metadata[meta_key]['station_id'] = station_id

                metadata[meta_key]["data_revision"] = self.data_revision
                metadata[meta_key]["variables"] = []
                metadata[meta_key]["var_info"] = {}
                # this is a list with indices of this station for each variable
                # not sure yet, if we really need that or if it speeds up things
                meta_idx[meta_key] = {}
                # last_station_id = station_id

                # Is floating point single value
                # time = stat.dtime[0] # LB: Check this
                for var in stat.vars_available:
                    if var not in data_obj.var_idx:
                        VAR_IDX += 1
                        data_obj.var_idx[var] = VAR_IDX

                    var_idx = data_obj.var_idx[var]

                    val = stat[var]
                    metadata[meta_key]["var_info"][var] = vi = {}
                    if isinstance(val, VerticalProfile):
                        altitude = val.altitude
                        data = val.data
                        add = np.prod(data.shape)
                        # err = val.data_err
                        metadata[meta_key]["var_info"]["altitude"] = via = {}

                        vi.update(val.var_info[var])
                        via.update(val.var_info["altitude"])
                    else:
                        add = 1
                        altitude = np.nan
                        data = val
                    vi.update(stat.var_info[var])
                    stop = idx + add
                    # check if size of data object needs to be extended
                    if stop >= data_obj._ROWNO:
                        # if totnum < data_obj._CHUNKSIZE, then the latter is used
                        data_obj.add_chunk(add)

                    # write common meta info for this station
                    data_obj._data[idx:stop, col_idx["latitude"]] = stat["station_coords"][
                        "latitude"
                    ]
                    data_obj._data[idx:stop, col_idx["longitude"]] = stat["station_coords"][
                        "longitude"
                    ]
                    data_obj._data[idx:stop, col_idx["altitude"]] = stat["station_coords"][
                        "altitude"
                    ]
                    data_obj._data[idx:stop, col_idx["meta"]] = meta_key

                    # write data to data object
                    data_obj._data[idx:stop, col_idx["time"]] = np.repeat(
                        stat.dtime, data.shape[-1]
                    )
                    data_obj._data[idx:stop, col_idx["stoptime"]] = np.repeat(
                        stat.dtime, data.shape[-1]
                    )
                    data_obj._data[idx:stop, col_idx["data"]] = data.flatten()
                    data_obj._data[idx:stop, col_idx["dataaltitude"]] = np.tile(
                        altitude, data.shape[0]
                    )
                    data_obj._data[idx:stop, col_idx["varidx"]] = var_idx

                    if var not in meta_idx[meta_key]:
                        meta_idx[meta_key][var] = []
                    meta_idx[meta_key][var].extend(list(range(idx, stop)))

                    if var not in metadata[meta_key]["variables"]:
                        metadata[meta_key]["variables"].append(var)

                    idx += add

            except Exception as e:
                self.read_failed.append(_file)
                logger.exception(f"Failed to read file {os.path.basename(_file)} (ERR: {repr(e)})")

        # shorten data_obj._data to the right number of points
        data_obj._data = data_obj._data[:idx]

        return data_obj

    def _get_exclude_filelist(self):  # pragma: no cover
        """Get list of filenames that are supposed to be ignored"""
        exclude = []

        files = files = (Path(self.data_dir) / "EXCLUDE").glob("*.txt")
        for i, file in enumerate(files):
            if os.path.basename(file) not in self.EXCLUDE_CASES:
                continue
            count = 0
            num = None
            indata = False
            with open(file) as f:
                for line in f:
                    if indata:
                        exclude.append(line.strip())
                        count += 1
                    elif "Number of" in line:
                        num = int(line.split(":")[1].strip())
                        indata = True

            if not count == num:
                raise EprofileFileError(
                    f"Number of excluded files in {file} does not match the number of files found in the file"
                )
        self.exclude_files = list(dict.fromkeys(exclude))
        return self.exclude_files

    @override
    def get_file_list(self) -> list[Path]:
        """Perform recursive file search for all input variables

        Note
        ----
        Overloaded implementation of base class, since for Earlinet, the
        paths are variable dependent

        Parameters
        ----------
        vars_to_retrieve : list
            list of variables to retrieve
        pattern : str, optional
            file name pattern applied to search

        Returns
        -------
        list
            list containing file paths
        """
        exclude_files = {Path(file) for file in self._get_exclude_filelist()}
        if self.data_dir is None:
            raise ValueError("No data directory set")
        logger.info("Fetching EPROFILE data files...")

        all_files = set(f for f in Path(self.data_dir).rglob(self._FILEMASK) if f.is_file())
        files = list(all_files - exclude_files)
        self.files = files
        return files



from importlib import metadata

import pytest
from packaging.version import Version

from pyaerocom import const

pytest_plugins = [
    "tests.fixtures.mscw_ctm",
    "tests.fixtures.tm5",
    "tests.fixtures.emep_ci",
    "tests.fixtures.ebas",
    "tests.fixtures.aeronet",
    "tests.fixtures.stations",
    "tests.fixtures.collocated_data",
    "tests.fixtures.aeroval.config",
    "tests.fixtures.cams2_83.config",
    "tests.fixtures.pyaro",
]

TEST_RTOL = 1e-4

# skipif marker that is True if no access to metno PPI is provided
# (some tests are skipped in this case)
lustre_unavail = pytest.mark.skipif(
    not const.has_access_lustre,
    reason="Skipping tests that require access to AEROCOM database on METNo servers",
)

lustre_avail = pytest.mark.skipif(
    const.has_access_lustre, reason="Skipping tests that will crash if lustre can be accessed."
)

etopo1_unavail = pytest.mark.skipif(
    not const.ETOPO1_AVAILABLE, reason="Skipping tests that require access to ETOPO1 data"
)


def __package_installed(name: str) -> bool:
    try:
        metadata.version(name)
    except ModuleNotFoundError:
        return False
    return True


# iris >= 3.2 corrected an error in iris.cube.Cube.intersection
# see https://github.com/metno/pyaerocom/issues/588
iris_version = metadata.version("scitools-iris")
need_iris_32 = pytest.mark.xfail(
    Version(iris_version) < Version("3.2"),
    reason=f"results are different with iris {iris_version} < 3.2",
    strict=True,
)



import pathlib
from copy import deepcopy

import aerovaldb
import aerovaldb.routes
import pytest

from pyaerocom import GriddedData
from pyaerocom.aeroval import EvalSetup
from pyaerocom.aeroval.modelmaps_engine import ModelMapsEngine
from pyaerocom.exceptions import ModelVarNotAvailable
from tests.fixtures.aeroval.cfg_test_exp1 import CFG


@pytest.fixture(scope="function")
def cfg(tmpdir: pathlib.Path):
    cfg = deepcopy(CFG)
    cfg["json_basedir"] = f"{tmpdir}/data"
    cfg["coldata_basedir"] = f"{tmpdir}/coldata"

    return cfg


def test__process_map_var(cfg: dict):
    stp = EvalSetup(**cfg)
    engine = ModelMapsEngine(stp)
    with pytest.raises(ModelVarNotAvailable) as excinfo:
        engine._process_contour_map_var("LOTOS", "concco", False)

    assert "Cannot read data for model LOTOS" in str(excinfo.value)


def test__run(caplog, cfg: dict):
    stp = EvalSetup(**cfg)
    engine = ModelMapsEngine(stp)
    engine.run(model_list=["TM5-AP3-CTRL"], var_list=["conco"])
    assert "no data for model TM5-AP3-CTRL, skipping" in caplog.text


def test__run_reanalysefalse(tmp_path, caplog, cfg: dict):
    """Test the case reanalyse_existing=False for plot type contour"""

    cfg["reanalyse_existing"] = False
    # modify the config so to have just one output map geojson file, for simplicity
    cfg["ts_type"] = "daily"
    cfg["periods"] = ["20100615"]
    cfg["main_freq"] = "daily"

    json_basedir = tmp_path / "data"
    cfg["json_basedir"] = json_basedir
    # create expected geojson output file (empty file is ok:
    # in the case reanalyse_existing=False content is not checked, only existence)
    with aerovaldb.open(f"json_files:{json_basedir}") as db:
        db.put_contour("", "test", "exp1", "od550aer", "TM5-AP3-CTRL", timestep="1277942400000")

    stp = EvalSetup(**cfg)
    engine = ModelMapsEngine(stp)
    engine.run(model_list=["TM5-AP3-CTRL"], var_list=["od550aer"])
    assert (
        "Skipping contour processing of od550aer_TM5-AP3-CTRL: data already exists" in caplog.text
    )


def test__run_working(cfg: dict):
    cfg["plot_types"] = ("contour", "overlay")
    stp = EvalSetup(**cfg)
    engine = ModelMapsEngine(stp)
    engine.run(model_list=["TM5-AP3-CTRL"], var_list=["od550aer"])
    contours = engine.exp_output.avdb.query(aerovaldb.routes.Route.CONTOUR_TIMESPLIT)
    overlays = engine.exp_output.avdb.query(aerovaldb.routes.Route.MAP_OVERLAY)
    assert len(contours) > 0
    assert len(overlays) > 0
    assert contours[0].meta["obsvar"] == "od550aer"
    assert contours[0].meta["model"] == "TM5-AP3-CTRL"


@pytest.mark.parametrize(
    "maps_freq, result",
    [("monthly", "monthly"), ("yearly", "yearly"), ("coarsest", "yearly")],
)
def test__get_maps_freq(maps_freq, result, cfg: dict):
    cfg["maps_freq"] = maps_freq
    stp = EvalSetup(**cfg)
    engine = ModelMapsEngine(stp)
    freq = engine._get_maps_freq()

    assert freq == result


@pytest.mark.parametrize(
    "maps_freq,result,ts_types",
    [
        ("monthly", "monthly", ["daily", "monthly", "yearly"]),
        ("yearly", "yearly", ["daily", "monthly", "yearly"]),
        ("coarsest", "yearly", ["daily", "monthly", "yearly"]),
        ("coarsest", "monthly", ["hourly", "daily", "monthly"]),
        ("coarsest", "daily", ["weekly", "daily"]),
    ],
)
def test__get_read_model_freq(maps_freq, result, ts_types, cfg: dict):
    cfg["maps_freq"] = maps_freq
    stp = EvalSetup(**cfg)
    engine = ModelMapsEngine(stp)
    freq = engine._get_read_model_freq(ts_types)

    assert freq == result


@pytest.mark.parametrize(
    "maps_freq,ts_types,errormsg",
    [
        (
            "daily",
            ["monthly", "yearly"],
            "Could not find any model data for given maps_freq.*",
        ),
    ],
)
def test__get_read_model_freq_error(maps_freq, ts_types, errormsg, cfg: dict):
    cfg["maps_freq"] = maps_freq
    stp = EvalSetup(**cfg)
    engine = ModelMapsEngine(stp)

    with pytest.raises(ValueError, match=errormsg):
        engine._get_read_model_freq(ts_types)


def test__read_model_data(cfg: dict):
    model_name = "TM5-AP3-CTRL"
    var_name = "od550aer"
    stp = EvalSetup(**cfg)
    engine = ModelMapsEngine(stp)

    data = engine._read_model_data(model_name, var_name)

    assert isinstance(data, GriddedData)



from __future__ import annotations

from pathlib import Path
from typing import Literal

import numpy as np
import pytest

from pyaerocom import const
from pyaerocom.aux_var_helpers import (
    compute_ac550dryaer,
    compute_ang4470dryaer_from_dry_scat,
    compute_sc440dryaer,
    compute_sc550dryaer,
    compute_sc700dryaer,
)
from pyaerocom.exceptions import (
    DataCoverageError,
    MetaDataError,
    NotInFileError,
    TemporalResolutionError,
    VariableDefinitionError,
    VarNotAvailableError,
)
from pyaerocom.io.ebas_nasa_ames import EbasNasaAmesFile
from pyaerocom.io.ebas_varinfo import EbasVarInfo
from pyaerocom.io.read_ebas import ReadEbas, ReadEbasOptions
from pyaerocom.stationdata import StationData
from pyaerocom.ungridded_data_metadata import UngriddedDataMetadata


@pytest.fixture(scope="module")
def reader() -> ReadEbas:
    return ReadEbas("EBASSubset")


def test_DATA_ID(reader: ReadEbas):
    assert reader.data_id == "EBASSubset"


def test_FILE_SUBDIR_NAME(reader: ReadEbas):
    assert reader.FILE_SUBDIR_NAME == "data"


def test_SQL_DB_NAME(reader: ReadEbas):
    # Name of sqlite database file
    assert reader.SQL_DB_NAME == "ebas_file_index.sqlite3"


def test_SUPPORTED_DATASETS(reader: ReadEbas):
    assert reader.SUPPORTED_DATASETS == ["EBASMC", "EBASSubset"]


def test_CACHE_SQLITE_FILE(reader: ReadEbas):
    assert reader.CACHE_SQLITE_FILE == ["EBASMC"]


def test_TS_TYPE(reader: ReadEbas):
    assert reader.TS_TYPE == "undefined"


def test_MERGE_STATIONS(reader: ReadEbas):
    assert reader.MERGE_STATIONS == {
        "Birkenes": "Birkenes II",
        "Rörvik": "Råö",
        "Vavihill": "Hallahus",
        "Virolahti II": "Virolahti III",
    }


def test_DEFAULT_VARS(reader: ReadEbas):
    assert reader.DEFAULT_VARS == reader.PROVIDES_VARIABLES


def test_TS_TYPE_CODES(reader: ReadEbas):
    assert reader.TS_TYPE_CODES == {
        "1mn": "minutely",
        "1h": "hourly",
        "1d": "daily",
        "1w": "weekly",
        "1mo": "monthly",
        "mn": "minutely",
        "h": "hourly",
        "d": "daily",
        "w": "weekly",
        "mo": "monthly",
    }


def test_AUX_REQUIRES(reader: ReadEbas):
    AUX_REQUIRES = dict(
        sc550dryaer=["sc550aer", "scrh"],
        sc440dryaer=["sc440aer", "scrh"],
        sc700dryaer=["sc700aer", "scrh"],
        ac550dryaer=["ac550aer", "acrh"],
        ang4470dryaer=["sc440dryaer", "sc700dryaer"],
    )
    for var_name, requires in AUX_REQUIRES.items():
        assert var_name in reader.AUX_REQUIRES
        assert requires == reader.AUX_REQUIRES[var_name]


def test_AUX_USE_META(reader: ReadEbas):
    AUX_USE_META = dict(
        sc550dryaer="sc550aer",
        sc440dryaer="sc440aer",
        sc700dryaer="sc700aer",
        ac550dryaer="ac550aer",
    )
    for var, ovar in AUX_USE_META.items():
        assert var in reader.AUX_USE_META
        assert reader.AUX_USE_META[var] == ovar


def test_AUX_FUNS(reader: ReadEbas):
    AUX_FUNS = dict(
        sc440dryaer=compute_sc440dryaer,
        sc550dryaer=compute_sc550dryaer,
        sc700dryaer=compute_sc700dryaer,
        ac550dryaer=compute_ac550dryaer,
        ang4470dryaer=compute_ang4470dryaer_from_dry_scat,
    )
    for var, func in AUX_FUNS.items():
        assert var in reader.AUX_FUNS
        assert reader.AUX_FUNS[var] == func


def test_ASSUME_AE_SHIFT_WVL(reader: ReadEbas):
    assert reader.ASSUME_AE_SHIFT_WVL == 1.0


def test_ASSUME_AAE_SHIFT_WVL(reader: ReadEbas):
    assert reader.ASSUME_AAE_SHIFT_WVL == 1.0


def test_IGNORE_FILES(reader: ReadEbas):
    assert reader.IGNORE_FILES == [
        "CA0420G.20100101000000.20190125102503.filter_absorption_photometer.aerosol_absorption_coefficient.aerosol.1y.1h.CA01L_Magee_AE31_ALT.CA01L_aethalometer.lev2.nas",
        "DK0022R.20180101070000.20191014000000.bulk_sampler..precip.1y.15d.DK01L_bs_22.DK01L_IC.lev2.nas",
        "DK0012R.20180101070000.20191014000000.bulk_sampler..precip.1y.15d.DK01L_bs_12.DK01L_IC.lev2.nas",
        "DK0008R.20180101070000.20191014000000.bulk_sampler..precip.1y.15d.DK01L_bs_08.DK01L_IC.lev2.nas",
        "DK0005R.20180101070000.20191014000000.bulk_sampler..precip.1y.15d.DK01L_bs_05.DK01L_IC.lev2.nas",
    ]


def test_opts(reader: ReadEbas):
    DEFAULTS = dict(
        prefer_statistics=["arithmetic mean", "median"],
        ignore_statistics=["percentile:15.87", "percentile:84.13"],
        wavelength_tol_nm=50,
        shift_wavelengths=True,
        assume_default_ae_if_unavail=True,
        check_correct_MAAP_wrong_wvl=False,
        eval_flags=True,
        keep_aux_vars=False,
        convert_units=True,
        try_convert_vmr_conc=True,
        ensure_correct_freq=False,
        freq_from_start_stop_meas=True,
        freq_min_cov=0.0,
    )
    opts = reader._opts
    assert isinstance(opts, dict)
    assert "default" in opts
    assert isinstance(opts["default"], ReadEbasOptions)
    assert len(DEFAULTS) == len(opts["default"])
    for opt, val in opts["default"].items():
        assert opt in DEFAULTS
        assert val == DEFAULTS[opt]


def test_data_id(reader: ReadEbas):
    assert reader.data_id == "EBASSubset"


def test_file_dir(reader: ReadEbas):
    assert reader._file_dir is None

    with pytest.raises(FileNotFoundError):
        reader.file_dir = 42

    path = Path(reader.file_dir)
    assert path.name == "data"
    assert path.exists()
    assert path.is_dir()

    reader.file_dir = str(path)  # sets private attr _file_dir
    assert reader._file_dir == reader.file_dir == str(path)


def test_FILE_REQUEST_OPTS(reader: ReadEbas):
    assert reader.FILE_REQUEST_OPTS == [
        "variables",
        "start_date",
        "stop_date",
        "station_names",
        "matrices",
        "altitude_range",
        "lon_range",
        "lat_range",
        "instrument_types",
        "statistics",
        "datalevel",
    ]


def test__FILEMASK(reader: ReadEbas):
    with pytest.raises(AttributeError) as e:
        reader._FILEMASK
    assert str(e.value) == (
        "Irrelevant for EBAS implementation, since SQL database is used for finding valid files"
    )


def test_NAN_VAL(reader: ReadEbas):
    with pytest.raises(AttributeError) as e:
        reader.NAN_VAL
    assert (
        str(e.value)
        == "Irrelevant for EBAS implementation: Info about invalid measurements is extracted from header of NASA Ames files for each variable individually "
    )


def test_PROVIDES_VARIABLES(reader: ReadEbas):
    PROVIDES_VARIABLES = {
        "DEFAULT",
        "concebc",
        "concca",
        "concmg",
        "conck",
        "sc550aer",
        "sc440aer",
        "sc700aer",
        "sc550dryaer",
        "sc440dryaer",
        "sc700dryaer",
        "sc550lt1aer",
        "bsc550aer",
        "ac550aer",
        "ac550dryaer",
        "ac550lt1aer",
        "bsc550dryaer",
        "scrh",
        "acrh",
        "ts",
        "concso4",
        "SO4ugSm3",
        "concso4pm10",
        "concso4pm25",
        "concso4pm1",
        "concso2",
        "vmrso2",
        "concpm10",
        "concpm25",
        "concpm1",
        "concso4t",
        "concso4c",
        "concbc",
        "conceqbc",
        "concCec",
        "concCecpm25",
        "conctc",
        "concoa",
        "concoc",
        "concCoc",
        "concCocpm25",
        "concss",
        "concsspm10",
        "concsspm25",
        "concnh3",
        "concNnh3",
        "concno3",
        "concNno3pm10",
        "concNno3pm25",
        "concNno3pm1",
        "concno3pm10",
        "concno3pm25",
        "concno3pm1",
        "concnh4",
        "concNnh4",
        "concNhno3",
        "concNtno3",
        "concNtnh",
        "concno",
        "concno2",
        "concNno2",
        "conchcho",
        "conco3",
        "concco",
        "vmro3",
        "vmro3max",
        "vmrco",
        "vmrno2",
        "vmrno",
        "vmrisop",
        "vmrhcho",
        "vmrglyoxal",
        "vmrc2h6",
        "vmrc2h4",
        "concglyoxal",
        "concprcpoxs",
        "concprcpoxn",
        "concprcprdn",
        "wetoxs",
        "wetrdn",
        "wetoxn",
        "pr",
        "prmm",
        "concnh4pm10",
        "concnh4pm25",
        "concnh4pm1",
        "concCocpm10",
        "concNno",
        "concCecpm10",
        "concSso2",
        "vmrnh3",
        "proxydryoxn",
        "proxywetpm25",
        "concss25",
        "concprcpno3",
        "concprcpso4",
        "concCoc25",
        "concom25",
        "concom1",
        "concprcpnh4",
        "concsscoarse",
        "proxydryhno3",
        "proxydryhono",
        "proxydryn2o5",
        "proxydrynh3",
        "proxydrynh4",
        "proxydryno2",
        "proxydryno2no2",
        "proxydryno3c",
        "proxydryno3f",
        "proxydryo3",
        "proxydryoxs",
        "proxydryss",
        "proxydryna",
        "proxydrypm10",
        "proxydrypm25",
        "proxydryrdn",
        "proxydryso2",
        "proxydryso4",
        "proxywethno3",
        "proxywethono",
        "proxywetn2o5",
        "proxywetnh3",
        "proxywetnh4",
        "proxywetno2",
        "proxywetno2no2",
        "proxywetno3c",
        "proxywetno3f",
        "proxyweto3",
        "proxywetoxn",
        "proxywetoxs",
        "proxywetpm10",
        "proxywetrdn",
        "proxywetso2",
        "proxywetso4",
        "vmrhno3",
        "vmrtp",
        "wetnh4",
        "wetno3",
        "wetso4",
        "wetna",
        "concprcpoxst",
        "wetoxsc",
        "concprcpoxsc",
        "wetoxst",
        "concprcpna",
        "wetso4pr",
        "wetrdnpr",
        "wetoxspr",
    }

    assert set(reader.PROVIDES_VARIABLES) == (PROVIDES_VARIABLES)


def test_sqlite_database_file(reader: ReadEbas):
    file = reader.sqlite_database_file
    assert Path(file).exists()
    assert file.endswith("ebas_file_index.sqlite3")


@pytest.mark.parametrize(
    "vars_to_retrieve,constraints",
    [
        ("vmrno", {}),
        ("sc550aer", {}),
        ("sc550dryaer", {}),
        ("sc550dryaer", {"station_names": "Jungfraujoch"}),
        ("ac550aer", {}),
        ("concpm10", {}),
        ("conco3", {}),
        (["sc550aer", "ac550aer", "concpm10", "conco3"], {}),
        (["sc550aer", "ac550aer", "concpm10", "conco3"], {"station_names": "*Kose*"}),
    ],
)
def test_get_file_list(
    reader: ReadEbas, vars_to_retrieve: list[str] | str, constraints: dict[str, str]
):
    files = reader.get_file_list(vars_to_retrieve, **constraints)
    assert isinstance(files, list)
    assert files


def test_get_file_list_error(reader: ReadEbas):
    with pytest.raises(FileNotFoundError) as e:
        reader.get_file_list("vmrno", station_names="xkcd")
    assert (
        str(e.value)
        == "No files could be found for ['vmrno'] and reading constraints {'station_names': 'xkcd'}."
    )


def test__merge_lists(reader: ReadEbas):
    vars_to_retrieve = ["sc550aer", "ac550aer"]
    files = reader.get_file_list(vars_to_retrieve)
    assert list(reader._lists_orig) == vars_to_retrieve
    assert reader._merge_lists(reader._lists_orig) == files


def test__find_station_matches(reader: ReadEbas):
    assert reader._find_station_matches("*fraujo*") == ["Jungfraujoch"]


@pytest.mark.parametrize(
    "val,exception,error",
    [
        pytest.param(
            "Bla",
            FileNotFoundError,
            "No EBAS data files could be found for stations Bla",
            id="station not found",
        ),
        pytest.param(
            42,
            ValueError,
            "Need list or string...",
            id="wrong pattern type",
        ),
    ],
)
def test__find_station_matches_error(
    reader: ReadEbas, val: Literal["Bla", 42], exception: type[Exception], error: str
):
    with pytest.raises(exception) as e:
        reader._find_station_matches(val)
    assert str(e.value) == error


@pytest.mark.parametrize(
    "vars_to_retrieve,result",
    [
        (None, None),
        (["sconco3"], ["conco3"]),
    ],
)
def test__precheck_vars_to_retrieve(
    reader: ReadEbas, vars_to_retrieve: list[str] | None, result: list[str] | None
):
    if result is None:
        result = reader.PROVIDES_VARIABLES
    assert reader._precheck_vars_to_retrieve(vars_to_retrieve) == result


def test_get_ebas_var(reader: ReadEbas):
    var_name = "sc550aer"
    info = reader.get_ebas_var(var_name)
    assert isinstance(info, EbasVarInfo)
    assert var_name in reader._loaded_ebas_vars


@pytest.mark.parametrize(
    "var_name,exception,error",
    [
        pytest.param(
            "blaaablub",
            VariableDefinitionError,
            "Error (VarCollection): input variable blaaablub is not supported",
            id="VariableDefinitionError",
        ),
        pytest.param(
            "abs550aer",
            VarNotAvailableError,
            "Variable abs550aer is not available in EBAS interface",
            id="VarNotAvailableError",
        ),
    ],
)
def test_get_ebas_var_error(
    reader: ReadEbas, var_name: str, exception: type[Exception], error: str
):
    with pytest.raises(exception) as e:
        reader.get_ebas_var(var_name)
    assert str(e.value) == error


def test__get_var_cols(reader: ReadEbas, loaded_nasa_ames_example: EbasNasaAmesFile):
    info = EbasVarInfo("sc550aer")
    cols = reader._get_var_cols(info, loaded_nasa_ames_example)
    assert cols == [14, 17, 20]


@pytest.mark.parametrize(
    "var,error",
    [
        ("ac550aer", "Variable ac550aer could not be found in file"),
        ("sc550dryaer", ""),
    ],
)
def test__get_var_cols_error(
    reader: ReadEbas, loaded_nasa_ames_example: EbasNasaAmesFile, var: str, error: str
):
    info = EbasVarInfo(var)
    with pytest.raises(NotInFileError) as e:
        reader._get_var_cols(info, loaded_nasa_ames_example)
    assert str(e.value) == error


def test_find_var_cols(reader: ReadEbas, loaded_nasa_ames_example: EbasNasaAmesFile):
    vars_to_read = ["sc550aer", "scrh", "ts"]
    columns = reader.find_var_cols(vars_to_read, loaded_nasa_ames_example)
    assert columns["sc550aer"] == 17
    assert columns["scrh"] == 3
    assert columns["ts"] == 4


@pytest.mark.parametrize(
    "ts_type,tol_percent,num_flagged",
    [
        ("hourly", 5, 0),
        ("daily", 5, 8760),
        ("hourly", 0, 5840),
    ],
)
def test__flag_incorrect_frequencies(
    monkeypatch: pytest.MonkeyPatch,
    reader: ReadEbas,
    loaded_nasa_ames_example: EbasNasaAmesFile,
    ts_type: str,
    tol_percent: int,
    num_flagged: int,
):
    station = StationData()
    station.start_meas = loaded_nasa_ames_example.start_meas
    station.stop_meas = loaded_nasa_ames_example.stop_meas
    station.var_info["bla"] = dict(units="1")
    station.bla = np.ones_like(loaded_nasa_ames_example.start_meas)
    station.ts_type = ts_type

    monkeypatch.setattr("pyaerocom.TsType.TOL_SECS_PERCENT", tol_percent)
    reader._flag_incorrect_frequencies(station)

    assert "bla" in station.data_flagged
    flagged = station.data_flagged["bla"]
    assert flagged.sum() == num_flagged


conco3_tower_var_info = {
    "conco3": {
        "name": "ozone",
        "units": "ug m-3",
        "tower_inlet_height": "50.0 m",
        "measurement_height": "50.0 m",
        "instrument_name": "uv_abs_kre_0050",
        "volume_std._temperature": "293.15 K",
        "volume_std._pressure": "1013.25 hPa",
        "detection_limit": "1.995 ug/m3",
        "comment": "Data converted on import into EBAS from 'nmol/mol' to 'ug/m3' at standard conditions (293.15 K, 1013.25 hPa), conversion factor 1.99534. Variable metadata detection limit converted.",
        "matrix": "air",
        "statistics": "arithmetic mean",
        "ts_type": "hourly",
    }
}
vmro3_tower_var_info = {
    "vmro3": {
        "name": "ozone",
        "units": "nmol mol-1",
        "tower_inlet_height": "50.0 m",
        "measurement_height": "50.0 m",
        "instrument_name": "uv_abs_kre_0050",
        "detection_limit": "1.0 nmol/mol",
        "comment": "Data converted on import into EBAS from 'nmol/mol' to 'ug/m3' at standard conditions (293.15 K, 1013.25 hPa), conversion factor 1.99534. Variable metadata detection limit converted.",
        "matrix": "air",
        "statistics": "arithmetic mean",
        "ts_type": "hourly",
    }
}


@pytest.mark.parametrize(
    "issue_files,vars_to_retrieve,check",
    [
        ("o3_tower", "vmro3", dict(var_info=vmro3_tower_var_info)),
        ("o3_tower", "conco3", dict(var_info=conco3_tower_var_info)),
        ("pm10_tstype", "concpm10", dict(ts_type="2daily")),
        ("Jungfraujoch", ["sc550aer"], dict(station_name="Jungfraujoch")),
    ],
)
def test_read_file(reader: ReadEbas, ebas_issue_files: Path, vars_to_retrieve: str, check: dict):
    data = reader.read_file(ebas_issue_files, vars_to_retrieve)
    assert isinstance(data, StationData)
    for key, val in check.items():
        if isinstance(val, dict):
            for subkey, subval in val.items():
                for subsubkey, subsubval in subval.items():
                    if subsubkey == "comment":
                        assert data[key][subkey][subsubkey].startswith(subsubval)
                    else:
                        assert data[key][subkey][subsubkey] == subsubval
        else:
            assert data[key] == val


@pytest.mark.parametrize(
    "issue_files,vars_to_retrieve,exception,error",
    [
        pytest.param(
            "pm10_colsel",
            "concpm10",
            ValueError,
            "failed to identify unique data column",
            id="repeated column",
        ),
        pytest.param(
            "o3_neg_dt",
            "conco3",
            TemporalResolutionError,
            "Nasa Ames file contains neg. meas periods...",
            id="negative period",
        ),
        pytest.param(
            "o3_tstype",
            "conco3",
            TemporalResolutionError,
            "Failed to derive correct sampling frequency in LT0015R.",
            id="wrong freq",
        ),
    ],
)
def test_read_file_error(
    reader: ReadEbas,
    ebas_issue_files: Path,
    vars_to_retrieve: str,
    exception: type[Exception],
    error: str,
):
    with pytest.raises(exception) as e:
        reader.read_file(ebas_issue_files, vars_to_retrieve)
    assert str(e.value).startswith(error)


def test__try_get_pt_conversion(reader: ReadEbas):
    fname = (
        "CH0001G.19910101000000.20181029122358.uv_abs.ozone.air.1y.1h.CH01L_TEI94C_1.CH01L_O3..nas"
    )
    path = Path(reader.file_dir) / fname
    assert path.exists()

    data = EbasNasaAmesFile(path)
    with pytest.raises(MetaDataError):
        reader._try_get_pt_conversion(data.meta)

    p, T = reader._try_get_pt_conversion(data.var_defs[2])
    assert p == 65300  # Pa
    assert T == 265.15  # K


@pytest.mark.parametrize(
    "vars_to_retrieve,file_vars,num_meta,num_stats",
    [
        ("concno2", "concno2", 2, 2),
        ("vmrno2", "vmrno2", 2, 2),
        ("vmrno2", ["vmrno2", "concno2"], 4, 4),
        ("conco3", "conco3", 4, 4),
        ("vmro3", "conco3", 4, 4),
        ("concpm10", "concpm10", 4, 4),
        ("sc550aer", None, 5, 4),
        ("sc550dryaer", "sc550dryaer", 5, 4),
        ("ac550aer", "ac550aer", 4, 4),
    ],
)
def test_read(
    reader: ReadEbas,
    vars_to_retrieve: list[str] | str,
    ebas_files: list[Path] | None,
    num_meta: int,
    num_stats: int,
):
    data = reader.read(vars_to_retrieve, files=ebas_files)
    assert isinstance(data, UngriddedDataMetadata)
    assert len(data.metadata) == num_meta
    assert len(data.unique_station_names) == num_stats

    if isinstance(vars_to_retrieve, str):
        vars_to_retrieve = [vars_to_retrieve]
    for var in vars_to_retrieve:
        for meta in data.metadata.values():
            assert meta["var_info"][var]["units"] == const.VARS[var].units


@pytest.mark.parametrize("file_vars", ["sc550dryaer"])
def test_read_error(reader: ReadEbas, ebas_files: list[Path]):
    with pytest.raises(DataCoverageError) as e:
        reader.read("ac550aer", files=ebas_files)
    assert str(e.value) == "UngriddedData object appears to be empty"



from pydantic import (
    BaseModel,
    ConfigDict,
    NonNegativeInt,
    PositiveInt,
)

from pyaerocom.aeroval.json_utils import set_float_serialization_precision


class StatisticsSetup(BaseModel, extra="allow"):
    """
    Setup options for statistical calculations

    Attributes
    ----------
    weighted_stats : bool
        if True, statistics are calculated using area weights,
        this is only relevant for gridded / gridded evaluations.
    annual_stats_constrained : bool
        if True, then only sites are considered that satisfy a potentially
        specified annual resampling constraint (see
        :attr:`pyaerocom.colocation.ColocationSetup.min_num_obs`). E.g.

        lets say you want to calculate statistics (bias,
        correlation, etc.) for monthly model / obs data for a given site and
        year. Lets further say, that there are only 8 valid months of data, and
        4 months are missing, so statistics will be calculated for that year
        based on 8 vs. 8 values. Now if
        :attr:`pyaerocom.colocation.ColocationSetup.min_num_obs` is
        specified in way that requires e.g. at least 9 valid months to
        represent the whole year, then this station will not be considered in
        case `annual_stats_constrained` is True, else it will. Defaults to
        False.
    stats_tseries_base_freq : str, optional
        The statistics Time Series display in AeroVal (under Overall Evaluation)
        is computed in intervals of a certain frequency, which is specified
        via :attr:`TimeSetup.main_freq` (defaults to monthly). That is,
        monthly colocated data is used as a basis to compute the statistics
        for each month (e.g. if you have 10 sites, then statistics will be
        computed based on 10 monthly values for each month of the timeseries,
        1 value for each site). `stats_tseries_base_freq` may be specified in
        case a higher resolution is supposed to be used as a basis to compute
        the timeseries in the resolution specified by
        :attr:`TimeSetup.main_freq` (e.g. if daily is specified here, then for
        the above example 310 values would be used - 31 for each site - to
        compute the statistics for a given month (in this case, a month with 31
        days, obviously).
    drop_stats: tuple, optional
        tuple of strings with names of statistics (as determined by keys in
        aeroval.glob_defaults.py's statistics_defaults) to not compute. For example,
        setting drop_stats = ("mb", "mab"), results in json files in hm/ts with
        entries which do not contain the mean bias and mean absolute bias,
        but the other statistics are preserved.
    stats_decimals: int, optional
        If provided, overwrites the decimals key in glod_defaults for the statistics, which has a default of 3.
        Setting this higher of lower changes the number of decimals shown on the Aeroval webpage.
    round_floats_precision: int, optional
        Sets the precision argument for the function `pyaerocom.aaeroval.json_utils:set_float_serialization_precision`


    Parameters
    ----------
    kwargs
        any of the supported attributes, e.g.
        `StatisticsSetup(annual_stats_constrained=True)`

    """

    # Pydantic ConfigDict
    model_config = ConfigDict(protected_namespaces=())
    # StatisticsSetup attributes
    MIN_NUM: PositiveInt = 1
    weighted_stats: bool = True
    annual_stats_constrained: bool = False

    # Trends config
    add_trends: bool = False  # Adding trend calculations, only trends over the average time series over stations in a region
    avg_over_trends: bool = (
        False  # Adds calculation of avg over trends of time series of stations in region
    )
    obs_min_yrs: NonNegativeInt = 0  # Removes stations with less than this number of years of valid data (a year with data points in all four seasons) Should in most cases be the same as stats_min_yrs
    stats_min_yrs: PositiveInt = obs_min_yrs  # Calculates trends if number of valid years are equal or more than this. Should in most cases be the same as obs_min_yrs
    sequential_yrs: bool = False  # Whether or not the min_yrs should be sequential

    stats_tseries_base_freq: str | None = None
    forecast_evaluation: bool = False
    forecast_days: PositiveInt = 4
    use_fairmode: bool = False
    use_diurnal: bool = False
    obs_only_stats: bool = False
    model_only_stats: bool = False
    drop_stats: tuple[str, ...] = ()
    stats_decimals: int | None = None
    round_floats_precision: int | None = None

    if round_floats_precision:
        set_float_serialization_precision(round_floats_precision)



"""
General helper methods for the pyaerocom library.
"""

from __future__ import annotations

import logging
import math as ma
from datetime import datetime

import iris
import iris.analysis
import iris.coords
import iris.cube
import numpy as np
import pandas as pd
import xarray as xr

from pyaerocom import const
from pyaerocom.exceptions import (
    DataCoverageError,
    DataDimensionError,
    LongitudeConstraintError,
    MetaDataError,
    ResamplingError,
    VariableDefinitionError,
)
from pyaerocom.units.datetime.time_config import (
    PANDAS_RESAMPLE_OFFSETS,
    TS_TYPE_DATETIME_CONV,
    TS_TYPE_TO_PANDAS_FREQ,
    TS_TYPE_TO_FREQ_NAME,
    TS_TYPE_TO_NUMPY_FREQ,
)
from pyaerocom.units.datetime import TsType, is_year, to_pandas_timestamp
from pyaerocom.variable_helpers import get_variable

from pyaerocom.units import Unit

logger = logging.getLogger(__name__)

NUM_KEYS_META = ["longitude", "latitude", "altitude"]

STR_TO_IRIS = dict(
    count=iris.analysis.COUNT,
    gmean=iris.analysis.GMEAN,
    hmean=iris.analysis.HMEAN,
    max=iris.analysis.MAX,
    mean=iris.analysis.MEAN,
    median=iris.analysis.MEDIAN,
    sum=iris.analysis.SUM,
    nearest=iris.analysis.Nearest,
    linear=iris.analysis.Linear,
    areaweighted=iris.analysis.AreaWeighted,
)


def varlist_aerocom(varlist):
    if isinstance(varlist, str):
        varlist = [varlist]
    elif not (isinstance(varlist, list) or isinstance(varlist, tuple)):
        raise ValueError("Need string or list or tuple")
    output = []
    for var in varlist:
        try:
            _var = const.VARS[var].var_name_aerocom
            if _var not in output:
                output.append(_var)
        except VariableDefinitionError as e:
            logger.warning(repr(e))
    if len(output) == 0:
        raise ValueError("None of the input variables appears to be valid")
    return output


def delete_all_coords_cube(cube, inplace=True):
    """Delete all coordinates of an iris cube

    Parameters
    ----------
    cube : iris.cube.Cube
        input cube that is supposed to be cleared of coordinates
    inplace : bool
        if True, then the coordinates are deleted in the input object, else in
        a copy of it

    Returns
    -------
    iris.cube.Cube
        input cube without coordinates
    """
    if not inplace:
        cube = cube.copy()

    for aux_fac in cube.aux_factories:
        cube.remove_aux_factory(aux_fac)

    for coord in cube.coords():
        cube.remove_coord(coord)
    return cube


def extract_latlon_dataarray(
    arr,
    lat,
    lon,
    lat_dimname=None,
    lon_dimname=None,
    method="nearest",
    new_index_name=None,
    check_domain=True,
):
    """Extract individual lat / lon coordinates from `DataArray`
    lon/lat can also be x/y coordinates if the `DataArray` has only projected axes.

    Parameters
    ----------
    arr : DataArray
        data (must contain lat and lon dimensions)
    lat : array or similar
        1D array containing latitude coordinates
    lon : array or similar
        1D array containing longitude coordinates
    lat_dimname : str, optional
        name of latitude dimension in input data (if None, it assumes standard
        name)
    lon_dimname : str, optional
        name of longitude dimension in input data (if None, it assumes standard
        name)
    method : str
        how to interpolate to input coordinates (defaults to nearest neighbour)
    new_index_name : str, optional
        name of flattend latlon dimension (defaults to latlon)
    check_domain : bool
        if True, lat/lon domain of datarray is checked and all input coordinates
        that are outside of the domain are ignored.

    Returns
    -------
    DataArray
        data at input coordinates
    """
    if lat_dimname is None:
        lat_dimname = "lat"
    if lon_dimname is None:
        lon_dimname = "lon"
    if lat_dimname not in arr.dims and lat_dimname == "lat":
        for alias in const.COORDINFO["lat"].aliases:
            if alias in arr.dims:
                lat_dimname = alias
                break
    if lon_dimname not in arr.dims and lon_dimname == "lon":
        for alias in const.COORDINFO["lon"].aliases:
            if alias in arr.dims:
                lon_dimname = alias
                break
    if isinstance(lat, str):
        lat = [lat]
    if isinstance(lon, str):
        lon = [lon]
    if check_domain:
        arr_lat = arr[lat_dimname].data
        arr_lon = arr[lon_dimname].data
        lat0, lat1 = arr_lat.min(), arr_lat.max()
        lon0, lon1 = arr_lon.min(), arr_lon.max()
        new_lat = []
        new_lon = []
        for x, y in zip(lat, lon):
            if (lat0 <= x <= lat1) and (lon0 <= y <= lon1):
                new_lat.append(x)
                new_lon.append(y)
        if len(new_lat) == 0 and len(new_lon) == 0:
            raise DataCoverageError("Coordinates not found in dataarray")
        lat, lon = new_lat, new_lon
    if new_index_name is None:
        new_index_name = "latlon"
    where = {
        lat_dimname: xr.DataArray(lat, dims=new_index_name),
        lon_dimname: xr.DataArray(lon, dims=new_index_name),
    }
    subset = arr.sel(where, method=method)
    subset.attrs["lat_dimname"] = lat_dimname
    subset.attrs["lon_dimname"] = lon_dimname
    return subset


def lists_to_tuple_list(*lists):
    """Convert input lists (of same length) into list of tuples

    e.g. input 2 lists of latitude and longitude coords, output one list
    with tuple coordinates at each index
    """
    return list(zip(*lists))


def tuple_list_to_lists(tuple_list):
    """Convert list with tuples (e.g. (lat, lon)) into multiple lists"""
    return list(map(list, zip(tuple_list)))


def make_dummy_cube_latlon(
    lat_res_deg: float = 2,
    lon_res_deg: float = 3,
    lat_range: list[float] | tuple[float, float] = (-90, 90),
    lon_range: list[float] | tuple[float, float] = (-180, 180),
):
    """Make an empty Cube with given latitude and longitude resolution

    Dimensions will be lat, lon

    Parameters
    ----------
    lat_res_deg : float or int
        latitude resolution of grid
    lon_res_deg : float or int
        longitude resolution of grid
    lat_range : tuple or list
        2-element list containing latitude range. If `None`, then `(-90, 90)`
        is used.
    lon_range : tuple or list
        2-element list containing longitude range. If `None`, then `(-180, 180)`
        is used.

    Returns
    -------
    Cube
        dummy cube in input resolution
    """

    # Accept lists for lat_range and lon_range, but make sure correct length
    assert len(lat_range) == len(lon_range) == 2

    lons = np.arange(
        lon_range[0] + (lon_res_deg / 2), lon_range[1] + (lon_res_deg / 2), lon_res_deg
    )
    lats = np.arange(
        lat_range[0] + (lat_res_deg / 2), lat_range[1] + (lat_res_deg / 2), lat_res_deg
    )

    lon_circ = check_coord_circular(lons, modulus=360)
    latdim = iris.coords.DimCoord(
        lats,
        var_name="lat",
        standard_name="latitude",
        circular=False,
        units=Unit("degrees"),
    )

    londim = iris.coords.DimCoord(
        lons,
        var_name="lon",
        standard_name="longitude",
        circular=lon_circ,
        units=Unit("degrees"),
    )

    latdim.guess_bounds()
    londim.guess_bounds()
    dummy = iris.cube.Cube(np.ones((len(lats), len(lons))))

    dummy.add_dim_coord(latdim, 0)
    dummy.add_dim_coord(londim, 1)
    dummy.var_name = "dummy_grid"

    return dummy


def check_coord_circular(coord_vals, modulus, rtol=1e-5):
    """Check circularity of coordinate

    Parameters
    ----------
    coord_vals : list or ndarray
        values of coordinate to be tested
    modulus : float or int
        modulus of coordinate (e.g. 360 for longitude)
    rtol : float
        relative tolerance

    Returns
    -------
    bool
        True if circularity is given, else False

    Raises
    ------
    ValueError
        if circularity is given and results in overlap (right end of input
        array is mapped to a value larger than the first one at the left end
        of the array)

    """

    if len(coord_vals) < 2:
        logger.warning(
            "Checking coordinate values for circularity "
            "failed since coord array has less than 2 values"
        )
        return False
    step = coord_vals[-1] - coord_vals[-2]
    tol = step * rtol
    diff = coord_vals[-1] - coord_vals[0] + step
    if diff - tol > modulus:
        raise ValueError(
            "Circularity is given but results in overlap (right "
            "end of input array is mapped to a value larger than "
            "the first one at the left end of the array)."
        )
    if abs(modulus - diff) > tol:
        return False
    return True


def numpy_to_cube(data, dims=None, var_name=None, units=None, **attrs):
    """Make a cube from a numpy array

    Parameters
    ----------
    data : ndarray
        input data
    dims : list, optional
        list of :class:`iris.coord.DimCoord` instances in order of dimensions
        of input data array (length of list and shapes of each of the
        coordinates must match dimensions of input data)
    var_name : str, optional
        name of variable
    units : str
        unit of variable
    **attrs
        additional attributes to be added to metadata

    Returns
    -------
    iris.cube.Cube

    Raises
    ------
    DataDimensionError
        if input `dims` is specified and results in conflict
    """
    if not isinstance(data, np.ndarray):
        raise ValueError("Invalid input, need numpy array")
    cube = iris.cube.Cube(data)

    cube.var_name = var_name
    cube.units = units

    sh = data.shape
    if dims is not None:
        if not len(dims) == data.ndim:
            raise DataDimensionError("Input number of dimensios must match array dimension number")
        for i, dim in enumerate(dims):
            if not isinstance(dim, iris.coords.DimCoord):
                raise ValueError("Need iris.DimCoord...")
            elif not len(dim.points) == sh[i]:
                raise DataDimensionError(
                    f"Length mismatch between {dim.var_name} dim ({len(dim.points)}) "
                    f"and array dimension {i} ({sh[i]})"
                )
            cube.add_dim_coord(dim, i)

    cube.attributes.update(attrs)
    return cube


def copy_coords_cube(to_cube, from_cube, inplace=True):
    """Copy all coordinates from one cube to another

    Requires the underlying data to be the same shape.

    Warning
    --------
    This operation will delete all existing coordinates and auxiliary
    coordinates and will then copy the ones from the input data object.
    No checks of any kind will be performed

    Parameters
    ----------
    to_cube
    other : GriddedData or Cube
        other data object (needs to be same shape as this object)

    Returns
    -------
    GriddedData
        data object containing coordinates from other object
    """
    if not all([isinstance(x, iris.cube.Cube) for x in [to_cube, from_cube]]):
        raise ValueError("Invalid input. Need instances of iris.cube.Cube class...")

    if not from_cube.shape == to_cube.shape:
        raise DataDimensionError("Cannot copy coordinates: shape mismatch")

    to_cube = delete_all_coords_cube(to_cube, inplace)

    for i, dim_coord in enumerate(from_cube.dim_coords):
        to_cube.add_dim_coord(dim_coord, i)

    for aux_coord, dim in from_cube._aux_coords_and_dims:
        to_cube.add_aux_coord(aux_coord, dim)

    for aux_fac in from_cube.aux_factories:
        to_cube.add_aux_factory(aux_fac)
    return to_cube


def get_standard_name(var_name):
    """Converts AeroCom variable name to CF standard name

    Also handles alias names for variables, etc. or strings corresponding to
    older conventions (e.g. names containing 3D).

    Parameters
    ----------
    var_name : str
        AeroCom variable name

    Returns
    -------
    str
        corresponding standard name
    """
    from pyaerocom import const

    return const.VARS[var_name].standard_name


def isnumeric(val):
    """Check if input value is numeric

    Parameters
    ----------
    val
        input value to be checked

    Returns
    -------
    bool
        True, if input value corresponds to a range, else False.
    """
    from numbers import Number

    if isinstance(val, Number):
        return True
    return False


def isrange(val):
    """Check if input value corresponds to a range

    Checks if input is list, or array or tuple with 2 entries, or alternatively
    a slice that has defined start and stop and has set step to None.

    Note
    ----
    No check is performed, whether first entry is smaller than second entry if
    all requirements for a range are fulfilled.

    Parameters
    ----------
    val
        input value to be checked

    Returns
    -------
    bool
        True, if input value corresponds to a range, else False.
    """
    if isinstance(val, list | np.ndarray | tuple):
        if len(val) == 2:
            return True
        return False
    elif isinstance(val, slice):
        if val.step is not None or val.start is None or val.stop is None:
            return False
        return True
    return False


def _check_stats_merge(statlist, var_name, pref_attr, fill_missing_nan):
    has_errs = False
    is_3d = []
    stats = []
    for stat in statlist:
        if var_name not in stat:
            raise DataCoverageError(f"All input stations must contain {var_name} data")
        elif pref_attr is not None and pref_attr not in stat:
            raise MetaDataError(
                f"Cannot sort station relevance by attribute {pref_attr}. "
                f"At least one of the input stations does not contain this attribute"
            )
        elif not isinstance(stat[var_name], pd.Series):
            stat._to_ts_helper(var_name)
        # this will raise MetaDataError or TemporalResolutionError if there is
        # an unresolvable issue with sampling frequency
        stat.get_var_ts_type(var_name)

        is_3d.append(stat.check_if_3d(var_name))

        if var_name in stat.data_err:
            has_errs = True

        stats.append(stat)
    if np.any(is_3d):
        if not np.all(is_3d):
            raise ValueError(
                "Merge error: some of the input stations contain "
                "altitude info (suggesting profile data), others "
                "not."
            )
        is_3d = True
    else:
        is_3d = False
    return (stats, is_3d, has_errs)


def _merge_stats_2d(
    stats,
    var_name,
    sort_by_largest,
    pref_attr,
    add_meta_keys,
    resample_how,
    min_num_obs,
):
    if pref_attr is not None:
        stats.sort(key=lambda s: s[pref_attr])
    else:
        stats.sort(key=lambda s: len(s[var_name].dropna()))

    if sort_by_largest:
        stats = stats[::-1]

    # remove first station from the list
    merged = stats.pop(0)
    for i, stat in enumerate(stats):
        merged.merge_other(
            stat,
            var_name,
            add_meta_keys=add_meta_keys,
            resample_how=resample_how,
            min_num_obs=min_num_obs,
        )
    return merged


def _merge_stats_3d(stats, var_name, add_meta_keys, has_errs):
    dtime = []
    for stat in stats:
        _t = stat[var_name].index.unique()
        if not len(_t) == 1:
            raise NotImplementedError(
                "So far, merging of profile data "
                "requires that profile values are "
                "sampled at the same time"
            )
        dtime.append(_t[0])
    tidx = pd.DatetimeIndex(dtime)

    # AeroCom default vertical grid
    vert_grid = const.make_default_vert_grid()
    _data = np.ones((len(vert_grid), len(tidx))) * np.nan
    if has_errs:
        _data_err = np.ones((len(vert_grid), len(tidx))) * np.nan

    for i, stat in enumerate(stats):
        if i == 0:
            merged = stat
        else:
            merged.merge_meta_same_station(stat, add_meta_keys=add_meta_keys)

        _data[:, i] = np.interp(vert_grid, stat["altitude"], stat[var_name].values)

        if has_errs:
            try:
                _data_err[:, i] = np.interp(vert_grid, stat["altitude"], stat.data_err[var_name])
            except Exception:
                pass
    _coords = {"time": tidx, "altitude": vert_grid}

    d = xr.DataArray(data=_data, coords=_coords, dims=["altitude", "time"], name=var_name)
    d = d.sortby("time")
    merged[var_name] = d
    merged.dtime = d.time
    merged.altitude = d.altitude
    return merged


def merge_station_data(
    stats,
    var_name,
    pref_attr=None,
    sort_by_largest=True,
    fill_missing_nan=True,
    add_meta_keys=None,
    resample_how=None,
    min_num_obs=None,
):
    """Merge multiple StationData objects (from one station) into one instance

    Note
    ----
    all input :class:`StationData` objects need to have same attributes
    ``station_name``, ``latitude``, ``longitude`` and ``altitude``

    Parameters
    ----------
    stats : list
        list containing :class:`StationData` objects (note: all of these
        objects must contain variable data for the specified input variable)
    var_name : str
        data variable name that is to be merged
    pref_attr
        optional argument that may be used to specify a metadata attribute
        that is available in all input :class:`StationData` objects and that
        is used to order the input stations by relevance. The associated values
        of this attribute need to be sortable (e.g. revision_date). This is
        only relevant in case overlaps occur. If unspecified the relevance of
        the stations is sorted based on the length of the associated data
        arrays.
    sort_by_largest : bool
        if True, the result from the sorting is inverted. E.g. if
        ``pref_attr`` is unspecified, then the stations will be sorted based on
        the length of the data vectors, starting with the shortest, ending with
        the longest. This sorting result will then be inverted, if
        ``sort_by_largest=True``, so that the longest time series get's highest
        importance. If, e.g. ``pref_attr='revision_date'``, then the stations
        are sorted by the associated revision date value, starting with the
        earliest, ending with the latest (which will also be inverted if
        this argument is set to True)
    fill_missing_nan : bool
        if True, the resulting time series is filled with NaNs. NOTE: this
        requires that information about the temporal resolution (ts_type) of
        the data is available in each of the StationData objects.
    add_meta_keys : str or list, optional
        additional non-standard metadata keys that are supposed to be
        considered for merging.
    resample_how : str or dict, optional
        in case input stations come in different frequencies they are merged
        to the lowest common freq. This parameter can be used to control, which
        aggregator(s) are to be used (e.g. mean, median).
    min_num_obs : str or dict, optional
        in case input stations come in different frequencies they are merged
        to the lowest common freq. This parameter can be used to control minimum
        number of observation constraints for the downsampling.

    Returns
    -------
    StationData
        merged data

    """
    if isinstance(var_name, list):
        if len(var_name) > 1:
            raise NotImplementedError("Merging of multivar data not yet possible")
        var_name = var_name[0]

    stats, is_3d, has_errs = _check_stats_merge(stats, var_name, pref_attr, fill_missing_nan)
    # ToDo: data_err is not handled at the moment for 2D data, needs r
    # revision and should be done in StationData.merge, also 3D vs 2D
    # should be handled by StationData directly...
    if is_3d:
        merged = _merge_stats_3d(stats, var_name, add_meta_keys, has_errs)
    else:
        merged = _merge_stats_2d(
            stats,
            var_name,
            sort_by_largest,
            pref_attr,
            add_meta_keys,
            resample_how,
            min_num_obs,
        )

    if fill_missing_nan:
        try:
            merged.insert_nans_timeseries(var_name)
        except Exception as e:
            logger.warning(
                f"Could not insert NaNs into timeseries of variable {var_name} "
                f"after merging stations. Reason: {repr(e)}"
            )

    merged["stat_merge_pref_attr"] = pref_attr
    return merged


def _get_pandas_freq_and_offset(freq: str) -> tuple[str, pd.Timedelta | None]:
    """Helper to convert resampling info"""
    if freq in TS_TYPE_TO_PANDAS_FREQ:
        freq = TS_TYPE_TO_PANDAS_FREQ[freq]
    offset = None
    if freq in PANDAS_RESAMPLE_OFFSETS:
        offset = PANDAS_RESAMPLE_OFFSETS[freq]
    return (freq, offset)


def make_datetime_index(start, stop, freq):
    """Make pandas.DatetimeIndex for input specs

    Note
    ----
    If input frequency is specified in `PANDAS_RESAMPLE_OFFSETS`, an offset
    will be added (e.g. 15 days for monthly data).

    Parameters
    ----------
    start
        start time. Preferably as :class:`pandas.Timestamp`, else it will be
        attempted to be converted.
    stop
        stop time. Preferably as :class:`pandas.Timestamp`, else it will be
        attempted to be converted.
    freq
        frequency of datetime index.

    Returns
    -------
    DatetimeIndex
    """
    if not isinstance(start, pd.Timestamp):
        start = to_pandas_timestamp(start)
    if not isinstance(stop, pd.Timestamp):
        stop = to_pandas_timestamp(stop)

    freq, offset = _get_pandas_freq_and_offset(freq)
    idx = pd.date_range(start=start, end=stop, freq=freq)
    if offset is not None:
        idx = idx + offset
    return idx


def make_datetimeindex_from_year(freq, year):
    """Create pandas datetime index

    Parameters
    ----------
    freq : str
        pandas frequency str
    year : int
        year

    Returns
    -------
    pandas.DatetimeIndex
        index object
    """
    start, stop = start_stop_from_year(year)
    return make_datetime_index(start, stop, freq)


def calc_climatology(s, start, stop, min_count=None, set_year=None, resample_how="mean"):
    """Compute climatological timeseries from pandas.Series

    Parameters
    ----------
    s : pandas.Series
        time series data
    start : numpy.datetime64 or similar
        start time of data used to compute climatology
    stop : numpy.datetime64 or similar
        start time of data used to compute climatology
    mincount_month : int, optional
        minimum number of observations required per aggregated month in
        climatological interval. Months not meeting this requirement will be
        set to NaN.
    set_year : int, optional
        if specified, the output data will be assigned the input year. Else
        the middle year of the climatological interval is used.
    resample_how : str
        string specifying how the climatological timeseries is to be
        aggregated

    Returns
    -------
    DataFrame
        dataframe containing climatological timeseries as
        well as columns std and count
    """
    if not isinstance(start, pd.Timestamp):
        start, stop = start_stop(start, stop)
    sc = s[start:stop]
    sc.dropna(inplace=True)

    if len(sc) == 0:
        raise ValueError(
            "Cropping input time series in climatological interval resulted in empty series"
        )
    if set_year is None:
        set_year = int(start.year + (stop.year - start.year) / 2) + 1

    df = pd.DataFrame(sc)
    df["month"] = df.index.month

    clim = df.groupby("month").agg([resample_how, "std", "count"])

    # clim.columns = clim.columns.droplevel(0)
    clim.columns = ["data", "std", "numobs"]
    idx = [np.datetime64(f"{set_year}-{x:02d}-15") for x in clim.index.values]
    clim.set_index(pd.DatetimeIndex(idx), inplace=True)
    if min_count is not None:
        mask = clim["numobs"] < min_count
        clim.loc[mask, "data"] = np.nan
    return clim


def resample_timeseries(ts, freq, how=None, min_num_obs=None):
    """Resample a timeseries (pandas.Series)

    Parameters
    ----------
    ts : Series
        time series instance
    freq : str
        new temporal resolution (can be pandas freq. string, or pyaerocom
        ts_type)
    how
        aggregator to be used, accepts everything that is accepted by
        :func:`pandas.core.resample.Resampler.agg` and in addition,
        percentiles may be provided as str using e.g. 75percentile as input for
        the 75% percentile.
    min_num_obs : int, optional
        minimum number of observations required per period (when downsampling).
        E.g. if input is in daily resolution and freq is monthly and
        min_num_obs is 10, then all months that have less than 10 days of data
        are set to nan.

    Returns
    -------
    Series
        resampled time series object
    """
    if how is None:
        how = "mean"
    elif "percentile" in how:
        p = int(how.split("percentile")[0])
        how = lambda x: np.nanpercentile(x, p)  # noqa: E731

    freq, offset = _get_pandas_freq_and_offset(freq)
    resampler = ts.resample(freq)

    data = resampler.agg(how)
    if min_num_obs is not None:
        numobs = resampler.count()
        # df = resampler.agg([how, 'count'])
        invalid = numobs < min_num_obs
        if np.any(invalid):
            data.values[invalid] = np.nan
    if offset is not None:
        data.index = data.index + offset
    return data


def resample_time_dataarray(arr, freq, how=None, min_num_obs=None):
    """Resample the time dimension of a :class:`xarray.DataArray`

    Note
    ----
    The dataarray must have a dimension coordinate named "time"

    Parameters
    ----------
    arr : DataArray
        data array to be resampled
    freq : str
        new temporal resolution (can be pandas freq. string, or pyaerocom
        ts_type)
    how : str
        how to aggregate (e.g. mean, median)
    min_num_obs : int, optional
        minimum number of observations required per period (when downsampling).
        E.g. if input is in daily resolution and freq is monthly and
        min_num_obs is 10, then all months that have less than 10 days of data
        are set to nan.

    Returns
    -------
    DataArray
        resampled data array object

    Raises
    ------
    IOError
        if data input `arr` is not an instance of :class:`DataArray`
    DataDimensionError
        if time dimension is not available in dataset
    """
    if how is None:
        how = "mean"
    elif "percentile" in how:
        raise NotImplementedError(
            "percentile based resampling is not yet available for xarray based data"
        )

    if not isinstance(arr, xr.DataArray):
        raise OSError(f"Invalid input for arr: need DataArray, got {type(arr)}")
    elif "time" not in arr.dims:
        raise DataDimensionError("Cannot resample time: input DataArray has no time dimension")

    to = TsType(freq)
    pd_freq = to.to_pandas_freq()
    invalid = None
    if min_num_obs is not None:
        invalid = arr.resample(time=pd_freq).count(dim="time") < min_num_obs

    freq, offset = _get_pandas_freq_and_offset(freq)
    resampler = arr.resample(time=pd_freq)

    try:
        aggfun = getattr(resampler, how)
    except AttributeError:
        raise ResamplingError(f"Invalid aggregator {how} for temporal resampling of DataArray...")
    arr = aggfun(dim="time")

    if offset is not None:
        arr["time"] = arr.get_index("time") + offset
    if invalid is not None:
        arr.data[invalid.data] = np.nan
    return arr


def same_meta_dict(meta1, meta2, ignore_keys=["PI"], num_keys=NUM_KEYS_META, num_rtol=1e-2):
    """Compare meta dictionaries

    Parameters
    ----------
    meta1 : dict
        meta dictionary that is to be compared with ``meta2``
    meta2 : dict
        meta dictionary that is to be compared with ``meta1``
    ignore_keys : list
        list containing meta keys that are supposed to be ignored
    num_keys : keys that contain numerical values
    num_rtol : float
        relative tolerance level for comparison of numerical values

    Returns
    -------
    bool
        True, if dictionaries are the same, else False
    """
    if not meta1.keys() == meta2.keys():
        return False
    for k, v in meta1.items():
        if k in ignore_keys:
            continue
        elif k in num_keys:
            if not ma.isclose(v, meta2[k], rel_tol=num_rtol):
                return False
        elif isinstance(v, dict):
            if not same_meta_dict(v, meta2[k]):
                return False
        else:
            if not v == meta2[k]:
                return False
    return True


def str_to_iris(key, **kwargs):
    """Mapping function that converts strings into iris analysis objects

    Please see dictionary ``STR_TO_IRIS`` in this module for valid definitions

    Parameters
    ----------
    key : str
        key of :attr:`STR_TO_IRIS` dictionary

    Returns
    -------
    obj
        corresponding iris analysis object (e.g. Aggregator, method)
    """
    key = key.lower()
    if key not in STR_TO_IRIS:
        raise KeyError(
            f"No iris.analysis object available for key {key}, please "
            f"choose from {STR_TO_IRIS.keys()}"
        )
    val = STR_TO_IRIS[key]
    if callable(val):
        return val(**kwargs)
    return val


def _check_climatology_timestamp(t):
    if isnumeric(t) and t == 9999:
        return pd.Timestamp("1-1-2222")
    elif isinstance(t, np.datetime64):
        tstr = str(t)
        if tstr.startswith("9999"):
            return pd.Timestamp(tstr.replace("9999", "2222"))
    elif isinstance(t, str) and "9999" in t:
        return pd.Timestamp(t.replace("9999", "2222"))
    elif isinstance(t, datetime) and t.year == 9999:
        return pd.Timestamp(t.replace(year=2222))
    raise ValueError(f"Failed to identify {t} as climatological timestamp...")


def start_stop(start, stop=None, stop_sub_sec=True):
    """Create pandas timestamps from input start / stop values

    Note
    ----
    If input suggests climatological data in AeroCom format (i.e. year=9999)
    then the year is converted to 2222 instead since pandas cannot handle
    year 9999.

    Parameters
    -----------
    start
        start time (any format that can be converted to pandas.Timestamp)
    stop
        stop time (any format that can be converted to pandas.Timestamp)
    stop_sub_sec : bool
        if True and if input for stop is a year (e.g. 2015) then one second
        is subtracted from stop timestamp (e.g. if input stop is
        2015 and denotes "until 2015", then for the returned stop timestamp
        one second will be subtracted, so it would be 31.12.2014 23:59:59).

    Returns
    -------
    pandas.Timestamp
        start timestamp
    pandas.Timestamp
        stop timestamp

    Raises
    ------
    ValueError
        if input cannot be converted to pandas timestamps
    """
    isclim = False
    try:
        start = to_pandas_timestamp(start)
    except pd.errors.OutOfBoundsDatetime:  # probably climatology
        start = _check_climatology_timestamp(start)
        isclim = True
    if stop is None:
        if isclim:
            yr = 2222
        else:
            yr = start.year
        stop = to_pandas_timestamp(f"{yr}-12-31 23:59:59")
    else:
        try:
            subt_sec = False
            if isnumeric(stop):
                subt_sec = True
            stop = to_pandas_timestamp(stop)
            if subt_sec and stop_sub_sec:
                stop = stop - pd.Timedelta(1, "s")
        except pd.errors.OutOfBoundsDatetime:
            stop = _check_climatology_timestamp(stop)
    return (start, stop)


def start_stop_str(start, stop=None, ts_type=None):
    conv = TS_TYPE_DATETIME_CONV[ts_type]
    if is_year(start) and stop is None:
        return str(start)
    start, stop = start_stop(start, stop)
    start_str = start.strftime(conv)
    stop_str = stop.strftime(conv)
    if stop_str != start_str:
        return f"{start_str}-{stop_str}"
    return start_str


def start_stop_from_year(year):
    """Create start / stop timestamp from year

    Parameters
    ----------
    year : int
        the year for which start / stop is to be instantiated

    Returns
    -------
    numpy.datetime64
        start datetime
    numpy.datetime64
        stop datetime
    """
    start = np.datetime64(f"{year}-01-01T00:00:00")
    stop = np.datetime64(f"{year}-12-31T23:59:59")
    return (start, stop)


def to_datestring_YYYYMMDD(value):
    """Convert input time to string with format YYYYMMDD

    Parameters
    ----------
    value
        input time, may be string, datetime, numpy.datetime64 or
        pandas.Timestamp

    Returns
    -------
    str
        input formatted to string YYYYMMDD

    Raises
    ------
    ValueError
        if input is not supported
    """
    if isinstance(value, str) and len(value) == 8:
        logger.info(
            "Input is already string containing 8 chars. Assuming it "
            "is in the right format and returning unchanged"
        )
        return value
    try:
        return to_pandas_timestamp(value).strftime("%Y%m%d")
    except Exception as e:
        raise ValueError(
            f"Invalid input, need str, datetime, numpy.datetime64 or pandas.Timestamp. "
            f"Error: {repr(e)}"
        )


def get_constraint(lon_range=None, lat_range=None, time_range=None, meridian_centre=True):
    """Function that creates an :class:`iris.Constraint` based on input

    Note
    ----
    Please be aware of the definition of the longitudes in your data when
    cropping within the longitude dimension. The longitudes in your data may be
    defined either from **-180 <= lon <= 180** (pyaerocom standard) or from
    **0 <= lon <= 360**. In the former case (-180 -> 180) you can leave the
    additional input parameter ``meridian_centre=True`` (default).

    Parameters
    ----------
    lon_range : :obj:`tuple`, optional
        2-element tuple containing longitude range for cropping
        Example input to crop around meridian: `lon_range=(-30, 30)`
    lat_range : :obj:`tuple`, optional
        2-element tuple containing latitude range for cropping.
    time_range : :obj:`tuple`, optional
        2-element tuple containing time range for cropping. Allowed data
        types for specifying the times are

            1. a combination of 2 :class:`pandas.Timestamp` instances or
            2. a combination of two strings that can be directly converted\
            into :class:`pandas.Timestamp` instances (e.g.\
            `time_range=("2010-1-1", "2012-1-1")`) or
            3. directly a combination of indices (:obj:`int`).
    meridian_centre : bool
        specifies the coordinate definition range of longitude array. If True,
        then -180 -> 180 is assumed, else 0 -> 360

    Returns
    -------
    iris.Constraint
        the combined constraint from all valid input parameters
    """
    constraints = []
    if lon_range is not None:
        constraints.append(get_lon_rng_constraint(*lon_range, meridian_centre))
    if lat_range is not None:
        constraints.append(get_lat_rng_constraint(*lat_range))
    if time_range is not None:
        constraints.append(get_time_rng_constraint(*time_range))
    if len(constraints) > 0:
        c = constraints[0]
        for cadd in constraints[1:]:
            c = c & cadd
    return c


def get_lat_rng_constraint(low, high):
    """Create latitude constraint based on input range

    Parameters
    ----------
    low : float or int
        lower latitude coordinate
    high : float or int
        upper latitude coordinate

    Returns
    -------
    iris.Constraint
        the corresponding iris.Constraint instance

    """
    return iris.Constraint(latitude=lambda v: low <= v <= high)


def get_lon_rng_constraint(low, high, meridian_centre=True):
    """Create longitude constraint based on input range

    Parameters
    ----------
    low : float or int
        left longitude coordinate
    high : float or int
        right longitude coordinate
    meridian_centre : bool
        specifies the coordinate definition range of longitude array of the
        data to be cropped. If True, then -180 -> 180 is assumed, else 0 -> 360

    Returns
    -------
    iris.Constraint
        the corresponding iris.Constraint instance

    Raises
    ------
    ValueError
        if first coordinate in lon_range equals or exceeds second
    LongitudeConstraintError
        if the input implies cropping over border of longitude array
        (e.g. 160 -> - 160 if -180 <= lon <= 180).
    """
    if low == high:
        raise ValueError("the specified values are equal")
    elif low > high:
        raise ValueError("Left coordinate must exceed right coordinate")
    if meridian_centre:
        low, high = (low + 180) % 360 - 180, (high + 180) % 360 - 180
    else:
        low, high = low % 360, high % 360
    if low > high:
        msg = "Cannot crop over right border of longitude range"
        raise LongitudeConstraintError(msg)
    return iris.Constraint(longitude=lambda v: low <= v <= high)


def get_time_rng_constraint(start, stop):
    """Create iris.Constraint for data extraction along time axis

    Parameters
    ----------
    start : :obj:`Timestamp` or :obj:` str`
        start time of desired subset. If string, it must be convertible
        into :class:`pandas.Timestamp` (e.g. "2012-1-1")
    stop : :obj:`Timestamp` or :obj:` str`
        start time of desired subset. If string, it must be convertible
        into :class:`pandas.Timestamp` (e.g. "2012-1-1")

    Returns
    -------
    iris.Constraint
        iris Constraint instance that can, e.g., be used as input for
        :func:`pyaerocom.griddeddata.GriddedData.extract`
    """
    if not isinstance(start, pd.Timestamp):
        start = pd.Timestamp(start)
    if not isinstance(stop, pd.Timestamp):
        stop = pd.Timestamp(stop)

    t_lower = iris.time.PartialDateTime(year=start.year, month=start.month, day=start.day)
    t_upper = iris.time.PartialDateTime(year=stop.year, month=stop.month, day=stop.day)

    return iris.Constraint(time=lambda cell: t_lower <= cell <= t_upper)


def get_max_period_range(periods):
    start = min([int(per.split("-")[0]) for per in periods])
    stop = max(int(per.split("-")[1]) if len(per.split("-")) > 1 else int(per) for per in periods)

    return start, stop


def make_dummy_cube(
    var_name: str,
    start_yr: int = 2000,
    stop_yr: int = 2020,
    freq: str = "daily",
    dtype=float,
) -> iris.cube.Cube:
    startstr = f"{TS_TYPE_TO_FREQ_NAME[freq]} since {start_yr}-01-01 00:00"

    if freq not in TS_TYPE_TO_PANDAS_FREQ.keys():
        raise ValueError(f"{freq} not a recognized frequency")

    start_str = f"{start_yr}-01-01 00:00"
    stop_str = f"{int(stop_yr)}-12-31 23:00"
    times = pd.date_range(start_str, stop_str, freq=TS_TYPE_TO_PANDAS_FREQ[freq])

    time_since_start = (times - times[0]) / np.timedelta64(1, TS_TYPE_TO_NUMPY_FREQ[freq])
    unit = get_variable(var_name).units

    lat_range = (-90, 90)
    lon_range = (-180, 180)
    lat_res_deg = 45
    lon_res_deg = 90
    time_unit = Unit(startstr, calendar="gregorian")

    lons = np.arange(
        lon_range[0] + (lon_res_deg / 2), lon_range[1] + (lon_res_deg / 2), lon_res_deg
    )
    lats = np.arange(
        lat_range[0] + (lat_res_deg / 2), lat_range[1] + (lat_res_deg / 2), lat_res_deg
    )

    latdim = iris.coords.DimCoord(
        lats,
        var_name="lat",
        standard_name="latitude",
        long_name="Center coordinates for latitudes",
        circular=False,
        units=Unit("degrees"),
    )

    londim = iris.coords.DimCoord(
        lons,
        var_name="lon",
        standard_name="longitude",
        long_name="Center coordinates for longitudes",
        circular=False,
        units=Unit("degrees"),
    )

    timedim = iris.coords.DimCoord(
        time_since_start,
        var_name="time",
        standard_name="time",
        long_name="Time",
        units=time_unit,
    )

    latdim.guess_bounds()
    londim.guess_bounds()
    dummy = iris.cube.Cube(np.ones((len(times), len(lats), len(lons))), units=unit)

    dummy.add_dim_coord(latdim, 1)
    dummy.add_dim_coord(londim, 2)
    dummy.add_dim_coord(timedim, 0)
    dummy.var_name = var_name
    dummy.ts_type = freq

    dummy.data = dummy.data.astype(dtype)
    for coord in dummy.coords():
        coord.points = coord.points.astype(dtype)
    return dummy



CFG = {
    "model_cfg": {
        "EMEP": {
            "model_id": "CAMS2-83.EMEP.day0.FC",
            "model_data_dir": "/lustre/storeB/project/fou/kl/CAMS2_83/model",
            "gridded_reader_id": {"model": "ReadCAMS2_83"},
            "model_kwargs": {"daterange": ["2025-03-01", "2025-03-01"]},
        }
    },
    "obs_cfg": {
        "EEA": {
            "obs_id": "CAMS2_83.NRT",
            "obs_vars": ["concno2", "concco", "conco3", "concso2", "concpm10", "concpm25"],
            "web_interface_name": "EEA-UTD",
            "obs_vert_type": "Surface",
            "read_opts_ungridded": {
                "files": [
                    "/lustre/storeB/project/fou/kl/CAMS2_83/obs/202502/obsmacc4verif_20250228.csv",
                    "/lustre/storeB/project/fou/kl/CAMS2_83/obs/202503/obsmacc4verif_20250301.csv",
                ],
                "force_caching": True,
            },
            "obs_filters": {
                "concno2": {
                    "latitude": [30, 82],
                    "longitude": [-30, 90],
                    "station_id": ["NO0042*"],
                    "negate": "station_id",
                },
                "concco": {
                    "latitude": [30, 82],
                    "longitude": [-30, 90],
                    "station_id": ["BETN*"],
                    "negate": "station_id",
                },
                "conco3": {
                    "latitude": [30, 82],
                    "longitude": [-30, 90],
                    "station_id": ["NO0042*"],
                    "negate": "station_id",
                },
                "concso2": {
                    "latitude": [30, 82],
                    "longitude": [-30, 90],
                    "station_id": "GR0001*",
                    "negate": "station_id",
                },
                "concpm10": {
                    "latitude": [30, 82],
                    "longitude": [-30, 90],
                    "station_id": ["NO0042*"],
                    "negate": "station_id",
                },
                "concpm25": {
                    "latitude": [30, 82],
                    "longitude": [-30, 90],
                    "station_id": ["NO0042*"],
                    "negate": "station_id",
                },
            },
        }
    },
    "proj_id": "cams2-83",
    "exp_id": "test",
    "exp_name": "Test",
    "exp_descr": "Evaluation using EEA-MF NRT obs.",
    "exp_pi": "<a href='https://atmosphere.copernicus.eu/help-and-support'>CAMS user support</a>",
    "public": True,
    "json_basedir": None,
    "coldata_basedir": None,
    "reanalyse_existing": True,
    "only_json": False,
    "only_colocation": True,
    "add_model_maps": False,
    "only_model_maps": False,
    "clear_existing_json": False,
    "raise_exceptions": False,
    "use_cams2_83": True,
    "filter_name": "ALL-wMOUNTAINS",
    "ts_type": "hourly",
    "map_zoom": "Europe",
    "freqs": ["hourly", "daily"],
    "periods": ["20250301-20250301"],
    "main_freq": "hourly",
    "add_seasons": False,
    "use_meteorological_seasons": True,
    "use_diurnal": False,
    "resample_how": {"vmro3": {"daily": {"hourly": "max"}}},
    "zeros_to_nan": True,
    "colocate_time": False,
    "obs_remove_outliers": False,
    "model_remove_outliers": False,
    "harmonise_units": True,
    "regions_how": "country",
    "annual_stats_constrained": False,
    "weighted_stats": False,
    "forecast_evaluation": True,
    "forecast_days": 1,
    "use_fairmode": False,
    "use_cams2_83_fairmode": True,
    "drop_stats": ("mb", "mab"),
    "var_order_menu": [
        "conco3",
        "conco3mda8",
        "concno2",
        "concpm10",
        "concpm25",
        "concso2",
        "concco",
    ],
    "min_num_obs": {"daily": {"hourly": 18}},
}



"""
Functions for computing the FAIRMODE statistics

FAIRMODE is the Forum for Air Quality Modeling, an initiative to bring together air quality modelers and users.
    - Promote and Support the use of models by EU Member States
    - Emphasis is on model application for air quality policy (monitoring, regulation, etc.)
    - Develop harmonized set of tools to test whether or a not a model is fit for a given purpose
    - CAMS has to make use of FAIRMODE diagrams

This module contains methods to compute the relevant FAIRMODE statistics.
"""

import numpy as np

from pyaerocom.units.datetime import TsType

SPECIES = dict(
    concno2=dict(UrRV=0.24, RV=200, alpha=0.2, freq=TsType("hourly")),
    conco3mda8=dict(UrRV=0.18, RV=120, alpha=0.79, freq=TsType("daily")),
    concpm10=dict(UrRV=0.28, RV=50, alpha=0.25, freq=TsType("daily")),
    concpm25=dict(UrRV=0.36, RV=25, alpha=0.5, freq=TsType("daily")),
)


def _RMSU(mean: float, std: float, spec: str) -> float:
    """RMSU is the Root Mean Squared Uncertainty associated with the uncertainty of the observations, U(O_i)."""

    if spec not in SPECIES:
        raise ValueError(f"Unsupported {spec=}")

    UrRV = SPECIES[spec]["UrRV"]
    RV = SPECIES[spec]["RV"]
    alpha = SPECIES[spec]["alpha"]

    in_sqrt = (1 - alpha**2) * (mean**2 + std**2) + alpha**2 * RV**2

    return UrRV * np.sqrt(in_sqrt)


def _fairmode_sign(mod_std: float, obs_std: float, R: float) -> float:
    if obs_std <= 0 or R >= 1:  # guard against sqrt(<0) or div0 errors
        return 1
    a = abs(mod_std - obs_std) / (obs_std * np.sqrt(2 * (1 - R)))
    return 1 if a >= 1 else -1


def _crms(mod_std: float, obs_std: float, R: float) -> float:
    """Returns the Centered Root Mean Squared Error"""
    return np.sqrt(mod_std**2 + obs_std**2 - 2 * mod_std * obs_std * R)


def _mqi(rms: float, rmsu: float, *, beta: float) -> float:
    """Model Quality Indicator (MQI). Pass beta=1 for `beta MQI`"""
    return rms / (rmsu * beta)


def fairmode_stats(obs_var: str, stats: dict, freq: str) -> dict:
    if obs_var not in SPECIES or np.isnan(list(stats.values())).any():
        return {}

    # compute only what it makes sense to compute
    if TsType(freq) != SPECIES[obs_var]["freq"]:
        return {}

    mean = stats["refdata_mean"]
    obs_std = stats["refdata_std"]
    mod_std = stats["data_std"]
    R = stats["R"]
    bias = stats["mb"]
    rms = stats["rms"]

    assert obs_std >= 0, f"negative {obs_std=}"
    assert mod_std >= 0, f"negative {mod_std=}"
    assert -1 <= R <= 1, f"out of range {R=}"

    crms = _crms(mod_std, obs_std, R)  # sqrt(rms ** 2 - bias ** 2)
    sign = _fairmode_sign(mod_std, obs_std, R)
    rmsu = _RMSU(mean, obs_std, obs_var)
    beta_mqi = _mqi(rms, rmsu, beta=1)

    # Check that fairmode staistics are computed as expected by checking MQI (Model Quality Indicator) for beta = 1
    assert np.isclose(
        rmsu * beta_mqi,
        np.sqrt((bias) ** 2 + (mod_std - obs_std) ** 2 + (2 * obs_std * mod_std * (1 - R))),
        rtol=1e-2,
    ), "failed MQI check"

    fairmode = dict(
        RMSU=rmsu,
        sign=sign,
        crms=crms,
        bias=bias,
        rms=rms,
        beta_mqi=beta_mqi,
        **SPECIES[obs_var],
    )
    return fairmode



import fnmatch
import logging
import sys
from collections.abc import Iterator
from copy import deepcopy

import numpy as np
import numpy.typing as npt
import pandas as pd
from pyaro.timeseries import Reader

from pyaerocom import const
from pyaerocom.dynamic_rec_array import DynamicRecArray
from pyaerocom.exceptions import (
    DataCoverageError,
    DataExtractionError,
    MetaDataError,
    VarNotAvailableError,
)
from pyaerocom.helpers import merge_station_data, start_stop
from pyaerocom.metastandards import STANDARD_META_KEYS
from pyaerocom.stationdata import StationData
from pyaerocom.ungridded_data_container import UngriddedDataContainer
from pyaerocom.ungridded_data_metadata import UngriddedDataMetadata
from pyaerocom.units.datetime import TsType
from pyaerocom.units.units_helpers import get_unit_conversion_fac

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)


class UngriddedDataStructured(UngriddedDataMetadata):
    """Class implementing UngriddedData in a numpy structured array"""

    #: version for caching, needs also updating when UngriddedDataMetadata has changed
    __version__ = "0.02"
    _merging_error_logged = False

    _dtype = [
        ("meta_id", "i4"),
        ("var_id", "i2"),
        ("start_time", "datetime64[s]"),
        ("end_time", "datetime64[s]"),
        ("data", "f"),  # data-value
        ("stdev", "f"),  # data-error
        (
            "dataaltitude",
            "i2",
        ),  # altitude of measurement (might be different from station)
        ("flag", "i2"),
    ]
    _nan_types = {
        "meta_id": np.iinfo("i4").min,
        "var_id": np.iinfo("i2").min,
        "start_time": np.datetime64("NaT"),
        "end_time": np.datetime64("NaT"),
        "data": np.nan,
        "stdev": np.nan,
        "dataaltitude": np.iinfo("i2").min,
        "flag": np.iinfo("i2").min,
    }

    def __init__(self, num_points: int = 100):
        super().__init__()  # initialize metadata
        self._dra = self._create_data_chunk(num_points)

        # filters applied
        self.filter_hist = {}

        self._is_vertical_profile = False

    @override
    def _new_from_meta_blocks(self, meta_ids: list, total: int = 100):
        # check for new variables of the stations
        new_var_idx = {}
        new_metadata = {}
        for meta_id in meta_ids:
            meta = self.metadata[meta_id]
            new_metadata[meta_id] = meta
            for var in meta["var_info"]:
                if var in self.ALLOWED_VERT_COORD_TYPES:
                    continue
                new_var_idx[var] = self.var_idx[var]

        midx = np.isin(self._dra.data["meta_id"], meta_ids)
        # data with the selected stations
        nd = self._dra.data[midx]
        size = len(nd)
        if size == 0:
            raise DataExtractionError("Filtering results in empty data object")

        new = self.__class__(size)
        new._dra.data = nd
        new.metadata = new_metadata
        new.var_idx = new_var_idx
        new.filter_hist = self.filter_hist
        new._is_vertical_profile = self._is_vertical_profile

        return new

    def _create_data_chunk(self, size):
        """create a datachunk of size and initialize it to _nan_type values"""
        data = DynamicRecArray(capacity=size, dtype=self._dtype)
        for k, v in self._nan_types.items():
            data._array[k] = v
        return data

    @override
    def copy(self):
        new = self.__class__()
        self._copy_metadata_to(new)
        new._dra = deepcopy(self._dra)
        return new

    @property
    @override
    def shape(self):
        return self._dra.data.shape

    @property
    @override
    def has_flag_data(self):
        return (self._dra.data["flag"] != self._nan_types["flag"]).any()

    @override
    def set_flags_nan(self, inplace=False):
        if not self.has_flag_data:
            raise AttributeError("Ungridded data object does not contain flagged data points")
        if inplace:
            obj = self
        else:
            obj = self.copy()

        obj._dra.data["flag"] = self._nan_types["flag"]
        obj._add_to_filter_history("set_flags_nan")
        return obj

    def to_station_data(
        self,
        meta_idx,
        vars_to_convert=None,
        start=None,
        stop=None,
        freq=None,
        ts_type_preferred=None,
        merge_if_multi=True,
        merge_pref_attr=None,
        merge_sort_by_largest=True,
        insert_nans=False,
        allow_wildcards_station_name=True,
        add_meta_keys=None,
        resample_how=None,
        min_num_obs=None,
    ):
        if isinstance(vars_to_convert, str):
            vars_to_convert = [vars_to_convert]
        elif vars_to_convert is None:
            vars_to_convert = self.contains_vars
            if len(vars_to_convert) == 0:
                raise DataCoverageError("UngriddedData object does not contain any variables")
        if start is None and stop is None:
            start = pd.Timestamp("1970")
            stop = pd.Timestamp("2200")
        else:
            start, stop = start_stop(start, stop)

        if isinstance(meta_idx, str):
            # user asks explicitly for station name, find all meta indices
            # that match this station
            meta_idx = self.find_station_meta_indices(meta_idx, allow_wildcards_station_name)
        if not isinstance(meta_idx, list):
            meta_idx = [meta_idx]

        stats = []
        start, stop = np.datetime64(start), np.datetime64(stop)

        if len(meta_idx) > 1:
            sub_idx = np.isin(self._dra.data["meta_id"], meta_idx)
            subdata = self._dra.data[sub_idx]
        else:
            subdata = self._dra.data
        for idx in meta_idx:
            try:
                stat = self._metablock_to_stationdata(
                    idx, vars_to_convert, start, stop, add_meta_keys, data=subdata
                )
                if ts_type_preferred is not None:
                    if "ts_type" in stat["var_info"][vars_to_convert[0]].keys():
                        if TsType(stat["var_info"][vars_to_convert[0]]["ts_type"]) < TsType(
                            ts_type_preferred
                        ):
                            continue
                    elif "ts_type" in stat.keys():
                        if TsType(stat["ts_type"]) < TsType(ts_type_preferred):
                            continue
                    else:
                        raise KeyError("Could not find ts_type in stat")
                stats.append(stat)
            except (VarNotAvailableError, DataCoverageError) as e:
                logger.debug(f"Skipping meta index {idx}. Reason: {repr(e)}")
        if merge_if_multi and len(stats) > 1:
            if len(vars_to_convert) > 1:
                raise NotImplementedError(
                    "Cannot yet merge multiple stations with multiple variables."
                )
            if merge_pref_attr is None:
                merge_pref_attr = self._try_infer_stat_merge_pref_attr(stats)
            merged = merge_station_data(
                stats,
                vars_to_convert,
                pref_attr=merge_pref_attr,
                sort_by_largest=merge_sort_by_largest,
                fill_missing_nan=False,
                resample_how=resample_how,
                min_num_obs=min_num_obs,
            )
            stats = [merged]

        stats_ok = []
        for stat in stats:
            for var in vars_to_convert:
                if var not in stat:
                    continue
                if freq is not None:
                    stat.resample_time(
                        var,
                        freq,
                        how=resample_how,
                        min_num_obs=min_num_obs,
                        inplace=True,
                    )
                elif insert_nans:
                    stat.insert_nans_timeseries(var)
                if var in stat.data_flagged:
                    if len(stat.data_flagged[var]) != len(stat.dtime):
                        del stat.data_flagged[var]
                        if not self._merging_error_logged:
                            self._merging_error_logged = True
                            logger.warning(
                                "merging of station-data objects introduced rubbish flags, removing"
                            )
                if var in stat.data_err:
                    if len(stat.data_err[var]) != len(stat.dtime):
                        del stat.data_err[var]
                        if not self._merging_error_logged:
                            self._merging_error_logged = True
                            logger.warning(
                                "merging of station-data objects introduced rubbish stddev, removing"
                            )
                if np.all(np.isnan(stat[var].values)):
                    stat = stat.remove_variable(var)
            if any([x in stat for x in vars_to_convert]):
                stats_ok.append(stat)

        if len(stats_ok) == 0:
            raise DataCoverageError(
                f"{vars_to_convert} data could not be retrieved "
                f"for meta index (or station name) {meta_idx}"
            )
        elif len(stats_ok) == 1:
            # return StationData object and not list
            return stats_ok[0]
        return stats_ok

    def _metablock_to_stationdata(
        self,
        meta_idx,
        vars_to_convert,
        start=None,
        stop=None,
        add_meta_keys=None,
        data=None,
    ):
        """Convert one metadata index to StationData (helper method)

        Data might be a pre-computed internal dataset.

        See :func:`to_station_data` for input parameters
        """
        if add_meta_keys is None:
            add_meta_keys = []
        elif isinstance(add_meta_keys, str):
            add_meta_keys = [add_meta_keys]

        sd = StationData()
        meta = self.metadata[meta_idx]

        # TODO: make sure in reading classes that data_revision is assigned
        # to each metadata block and not only in self.data_revision
        rev = None
        if "data_revision" in meta:
            rev = meta["data_revision"]
        else:
            try:
                rev = self.get_data_revision(meta["data_id"])
            except Exception:
                logger.debug("Data revision could not be accessed")
        sd.data_revision = rev
        try:
            vars_avail = list(meta["var_info"])
        except KeyError:
            if "variables" not in meta or meta["variables"] in (None, []):
                raise VarNotAvailableError("Metablock does not contain variable information")
            vars_avail = meta["variables"]

        for key in STANDARD_META_KEYS + add_meta_keys:
            if key in sd.PROTECTED_KEYS:
                logger.warning(f"skipping protected key: {key}")
                continue
            try:
                sd[key] = meta[key]
            except KeyError:
                pass

        try:
            sd["ts_type_src"] = meta["ts_type"]
        except KeyError:
            pass

        # assign station coordinates explicitly
        for ck in sd.STANDARD_COORD_KEYS:
            try:
                sd.station_coords[ck] = meta[ck]
            except KeyError:
                pass
        # if no input variables are provided, use the ones that are available
        # for this metadata block
        if vars_to_convert is None:
            vars_to_convert = vars_avail

        # find overlapping variables (ignore all other ones)
        vars_avail = np.intersect1d(vars_to_convert, vars_avail)
        if not len(vars_avail) >= 1:
            raise VarNotAvailableError(
                "None of the input variables matches, or station does not contain data."
            )
        # init helper boolean that is set to True if valid data can be found
        # for at least one of the input variables
        if data is None:
            data = self._dra.data
        FOUND_ONE = False
        for var in vars_avail:
            # get indices of this station and variable
            idx = (data["meta_id"] == meta_idx) & (data["var_id"] == self.var_idx[var])

            # get subset
            subset = data[idx]

            # vector of timestamps corresponding to this variable
            dtime = subset["start_time"]

            # make sure to extract only valid timestamps
            if start is None:
                start = dtime.min()
            if stop is None:
                stop = dtime.max()

            # create access mask for valid time stamps
            tmask = (dtime >= start) & (dtime <= stop)

            # make sure there is some valid data
            if tmask.sum() == 0:
                logger.debug(
                    f"Ignoring station {sd['station_name']}, var {var} ({sd['data_id']}): "
                    f"no data available in specified time interval {start} - {stop}"
                )
                continue

            dtime = dtime[tmask]
            subset = subset[tmask]

            vals = subset["data"]
            if np.all(np.isnan(vals)):
                logger.debug(
                    f"Ignoring station {sd['station_name']}, var {var} ({sd['data_id']}): "
                    f"All values are NaN"
                )
                continue
            vals_err = subset["stdev"]
            flag_mask = subset["flag"] == self._nan_types["flag"]
            flagged = subset["flag"].astype("f4")
            flagged[flag_mask] = np.nan
            alt_mask = subset["dataaltitude"] == self._nan_types["dataaltitude"]
            altitude = subset["dataaltitude"].astype("f4")
            altitude[alt_mask] = np.nan

            series = pd.Series(vals, dtime)
            if not series.index.is_monotonic_increasing:
                series = series.sort_index()
            if any(~np.isnan(vals_err)):
                sd.data_err[var] = vals_err
            if any(~np.isnan(flagged)):
                sd.data_flagged[var] = flagged

            sd["dtime"] = series.index.values
            sd[var] = series
            sd["var_info"][var] = {}
            FOUND_ONE = True
            # check if there is information about altitude (then relevant 3D
            # variables and parameters are included too)
            if "var_info" in meta:
                vi = meta["var_info"]
            else:
                vi = {}
            assert isinstance(vi, dict)
            if not np.isnan(altitude).all():
                if "altitude" in vi:
                    sd.var_info["altitude"] = vi["altitude"]
                sd.altitude = altitude[0]  # TODO: Revise in case of moving stations
            if var in vi:
                sd.var_info[var].update(vi[var])

            if len(series.index) == len(series.index.unique()):
                sd.var_info[var]["overlap"] = False
            else:
                sd.var_info[var]["overlap"] = True
        if not FOUND_ONE:
            raise DataCoverageError(
                f"Could not retrieve any valid data for station {sd['station_name']} "
                f"and input variables {vars_to_convert}"
            )
        return sd

    @override
    def _generate_station_index(self, by_station_name=True, ignore_index=None):
        """Generates index to loop over station names or metadata block indices.
        Needs to be implemented for :func:`to_station_data_all` to work"""
        if ignore_index is None:
            if by_station_name:
                return self.unique_station_names  # all station names
            return list(range(len(self.metadata)))  # all meta indices

        if not by_station_name:
            from pyaerocom.helpers import isnumeric

            if isnumeric(ignore_index):
                ignore_index = [ignore_index]
            if not isinstance(ignore_index, list):
                raise ValueError("Invalid input for ignore_index, need number or list")
            return [i for i in range(len(self.metadata)) if i not in ignore_index]

        # by station name and ignore certation stations
        _iter = []
        if isinstance(ignore_index, str):
            ignore_index = [ignore_index]
        if not isinstance(ignore_index, list):
            raise ValueError("Invalid input for ignore_index, need str or list")
        for stat_name in self.unique_station_names:
            ok = True
            for name_or_pattern in ignore_index:
                if fnmatch.fnmatch(stat_name, name_or_pattern):
                    ok = False
            if ok:
                _iter.append(stat_name)
        return _iter

    def check_convert_var_units(self, var_name, to_unit=None, inplace=True):
        obj = self if inplace else self.copy()

        # get the unit
        if to_unit is None:
            to_unit = const.VARS[var_name]["units"]

        for meta_idx, meta in obj.metadata.items():
            if var_name in meta["var_info"]:
                try:
                    unit = meta["var_info"][var_name]["units"]
                except KeyError:
                    add_str = ""
                    if "unit" in meta["var_info"][var_name]:
                        add_str = (
                            "Corresponding var_info dict contains "
                            'attr. "unit", which is deprecated, please '
                            "check corresponding reading routine. "
                        )
                    raise MetaDataError(
                        f"Failed to access unit information for variable {var_name} "
                        f"in metadata block {meta_idx}. {add_str}"
                    )
                fac = get_unit_conversion_fac(unit, to_unit, var_name)
                if fac != 1:
                    idx = (obj._dra.data["meta_id"] == meta_idx) & (
                        obj._dra.data["var_id"] == self.var_idx[var_name]
                    )
                    obj._dra.data["data"][idx] *= fac
                meta["var_info"][var_name]["units"] = to_unit

        return obj

    @override
    def remove_outliers(
        self,
        var_name,
        inplace=False,
        low=None,
        high=None,
        unit_ref=None,
        move_to_trash=True,
    ):
        """see super.remove_outliers move_to_trash is ignored for now"""
        if inplace:
            new = self
        else:
            new = self.copy()

        new.check_convert_var_units(var_name, to_unit=unit_ref)

        if low is None:
            low = const.VARS[var_name].minimum
            logger.info(f"Setting {var_name} outlier lower lim: {low:.2f}")
        if high is None:
            high = const.VARS[var_name].maximum
            logger.info(f"Setting {var_name} outlier upper lim: {high:.2f}")
        var_idx = new.var_idx[var_name]
        var_mask = new._dra.data["var_id"] == var_idx

        all_data = new._dra.data["data"]
        invalid_mask = np.logical_or(all_data < low, all_data > high)

        mask = invalid_mask * var_mask
        invalid_vals = new._dra.data["data"][mask]
        new._dra.data["data"][mask] = np.nan

        if move_to_trash:
            logger.warning("trash not implemented")

        new._add_to_filter_history(
            f"Removed {len(invalid_vals)} outliers from {var_name} data "
            f"(range: {low}-{high}, in trash: {move_to_trash})"
        )
        return new

    @override
    def _len_datapoints(self, meta_idx, var):
        if isinstance(meta_idx, float):
            meta_idx = [meta_idx]
        if isinstance(var, str):
            var = [var]
        var_idx = [self.var_idx[v] for v in var]
        return np.sum(
            np.isin(self._dra.data["meta_id"], meta_idx)
            & np.isin(self._dra.data["var_id"], var_idx)
        )

    @override
    def extract_vars(self, var_names, check_index=True):
        var_names_unaliased = []
        for var_name in var_names:
            if var_name in self.contains_vars:
                var_names_unaliased.append(var_name)
            else:
                # try alias
                _var = const.VARS[var_name].var_name_aerocom
                if _var in self.contains_vars:
                    var_names_unaliased.append(_var)
                else:
                    raise VarNotAvailableError(f"No such variable {var_name} in data")

        var_ids = [self.var_idx[x] for x in var_names_unaliased]
        idx = np.isin(self._dra.data["var_id"], var_ids)
        new = self.__class__()
        new._dra.data = deepcopy(self._dra.data[idx])
        # fix the metadata
        self._copy_metadata_to(new)
        for i, var in enumerate(var_names):
            new.var_idx[var] = var_ids[i]
        new.metadata = {}
        for meta_id, meta in self.metadata.items():
            common_vars = [var for var in var_names if var in meta["var_info"]]
            if len(common_vars):
                new.metadata[meta_id] = deepcopy(meta)
                new.metadata[meta_id]["var_info"] = {}
                new.metadata[meta_id]["variables"] = common_vars
                for var in common_vars:
                    new.metadata[meta_id]["var_info"][var] = deepcopy(meta["var_info"][var])
        return new

    @override
    def extract_var(self, var_name, check_index=True):
        new = self.extract_vars([var_name], check_index)
        new.filter_hist.update(self.filter_hist)
        new._add_to_filter_history(
            f"Created {var_name} single var object from multivar UngriddedData instance"
        )
        return new

    @override
    def all_datapoints_var(self, var_name):
        return self.extract_var(var_name)._dra.data["data"]

    @override
    def append_station_data(
        self,
        stats: StationData | Iterator[StationData],
        add_meta_keys: list[str] = [],
    ):
        if isinstance(stats, StationData):
            stats = [stats]
        # last meta_idx
        meta_idx = np.max(self._dra._array["meta_id"])
        if meta_idx == self._nan_types["meta_id"]:
            meta_idx = -1  # start at 0
        for station_data in stats:
            meta_idx += 1
            if not isinstance(station_data, StationData):
                raise ValueError("Need instances of StationData or dicts")

            # each file is a metadata-set of its own
            self.metadata[meta_idx] = {}
            self.metadata[meta_idx].update(
                station_data.get_meta(
                    force_single_value=False, quality_check=False, add_none_vals=True
                )
            )
            for key in add_meta_keys:
                if key in station_data:
                    self.metadata[meta_idx][key] = station_data[key]
            contains_vars = list(station_data.var_info)
            self.metadata[meta_idx]["variables"] = contains_vars
            if "data_revision" in station_data:
                self.metadata[meta_idx]["data_revision"] = station_data.data_revision

            for var in contains_vars:
                vardata = station_data[var]
                if isinstance(vardata, pd.Series):
                    times = vardata.index
                    values = vardata.values
                else:
                    times = station_data["dtime"]
                    values = vardata
                    if not len(times) == len(values):
                        raise ValueError
                if var not in self.var_idx:
                    self.var_idx[var] = len(self.var_idx)
                var_idx = self.var_idx[var]
                self.metadata[meta_idx]["var_info"] = {}
                self.metadata[meta_idx]["var_info"][var] = {}

                self.metadata[meta_idx]["var_info"][var].update(station_data["var_info"][var])
                for x in ("longitude", "latitude", "altitude"):
                    if x not in self.metadata[meta_idx]["var_info"][var]:
                        self.metadata[meta_idx]["var_info"][var][x] = station_data[x]

                uds = UngriddedDataStructured(num_points=len(values))
                v_data = uds._dra._array  # access to raw numpy-array
                v_data["meta_id"][:] = meta_idx
                v_data["var_id"][:] = var_idx
                v_data["data"] = values
                v_data["start_time"] = times
                # v_data["end_time"] not used
                # v_data["dataaltitude"] not used
                if var in station_data.data_err:
                    v_data["stdev"] = station_data.data_err[var]
                if var in station_data.data_flagged:
                    flags = station_data.data_flagged[var]
                    v_data["flag"][:] = flags
                    nans = ~np.isfinite(flags)
                    v_data["flag"][nans] = UngriddedDataStructured._nan_types["flag"]
                self._dra.append(v_data)

    @staticmethod
    def from_pyaro(
        data_id: str, reader: Reader, vars_to_retrieve: list[str], **kwargs
    ) -> UngriddedDataContainer:
        """Convert data from a pyaro-reader to UngriddedDataStructured

        :param data_id: data_id identifier
        :param reader: pyaro-reader
        :param vars_to_retrieve: selection of variables
        :param kwargs: internal parameters for testing/benchmarking
        :return: data as UngriddedDataContainer
        """

        def _calculate_ts_type(
            start: npt.NDArray[np.datetime64], end: npt.NDArray[np.datetime64]
        ) -> npt.NDArray:
            """convert start and end-time arrays to a ts-type array

            :param start: start-times
            :param end: end-times
            :return: ts-types as string-array
            """
            seconds = (end - start).astype("timedelta64[s]").astype(np.int32)

            # below line is the same as
            # uniq_seconds = np.sort(np.unique(seconds))
            # but several orders of magnitude faster for large array with only few distinct elements
            uniq_seconds = np.sort(np.nonzero(np.bincount(seconds))[0])

            @np.vectorize(otypes=[str])
            def memoized_ts_type(x: np.int32) -> str:
                if x == 0:
                    return TsType("hourly")
                return str(TsType.from_total_seconds(x))

            uniq_tstypes = memoized_ts_type(uniq_seconds)

            # Use np.searchsorted to find indices of structured_array elements in sorted_keys
            indices = np.searchsorted(uniq_seconds, seconds)

            # Use np.take to map indices to values
            return np.take(uniq_tstypes, indices)

        def _station_tstype_to_int_array(
            sarray: np.ndarray, mapping: dict[tuple[str, str], int]
        ) -> np.ndarray:
            """converter an array-view consisting of two str columns ("stations", "tstype") to an
            int-array using a mapping

            :param sarray: station,tstype view of a structured array
            :param mapping: (station,tstype) -> int mapping dictionary
            :return: array of ints
            """
            keys = np.array(
                list(mapping.keys()),
                dtype=[
                    ("stations", sarray["stations"].dtype),
                    ("tstype", sarray["tstype"].dtype),
                ],
            )
            values = np.array(list(mapping.values()), dtype="i4")

            # Sort keys to ensure correct indexing
            sorted_indices = np.argsort(keys)
            sorted_keys = keys[sorted_indices]
            sorted_values = values[sorted_indices]

            # Use np.searchsorted to find indices of structured_array elements in sorted_keys
            indices = np.searchsorted(sorted_keys, sarray)

            # Use np.take to map indices to values
            return np.take(sorted_values, indices)

        class _VariableMetaIds:
            """Class containing for each variable a dictionary of tuples of station and ts_type to
            the corresponding meta-d
            """

            def __init__(self):
                self._counter: int = 0
                # a meta must be split by variable as well as station and tstype
                # dictionary about var_meta[var][(station,ts_type)] = meta_id
                self._mapping: dict[str, dict[tuple[str, str], int]] = {}

            def append_var_station_tstype(
                self,
                var: str,
                station_tstype: np.ndarray | None,
            ):
                """append values of an array containing station and ts_type tuples
                :param var: variable
                :param station_tstype: structured array of stations and tstype
                """
                if var not in self._mapping:
                    self._mapping[var] = {}
                if len(station_tstype) == 0:
                    return

                mapping = self._mapping[var]
                # below line is for EEA-data about 10x faster than
                # uarray = np.unique(station_tstype, axis=0)
                uarray = np.array(list(set(station_tstype.tolist())), dtype=station_tstype.dtype)

                for row in uarray:
                    sx = (row[0], row[1])
                    if sx not in mapping:
                        mapping[sx] = self._counter
                        self._counter += 1
                return

            def __getitem__(self, var) -> dict[tuple[str, str], int]:
                """Get the (stations, tstype) -> metaid dictionary

                :param var: variable name
                :return: dictionary of tuple of stations and tstype to meta_ids
                """
                return self._mapping[var]

        ugs = UngriddedDataStructured()
        ugs.var_idx = {var: i for i, var in enumerate(vars_to_retrieve)}

        # a meta must be split by variable as well as station and tstype
        var_metas = _VariableMetaIds()
        # unit dictionary, var_units[var] = unit
        var_units: dict[str, str] = {}
        for var in vars_to_retrieve:
            logger.info(f"Getting data of {var} from pyaro/{data_id}")
            if "bench_dataset" in kwargs:
                var_data = kwargs["bench_dataset"]
            else:
                var_data = reader.data(varname=var)
            logger.info(f"Converting data of {var} from pyaro/{data_id} to ungridded")
            tstype = _calculate_ts_type(start=var_data.start_times, end=var_data.end_times)
            stations = var_data.stations
            station_tstype = np.rec.array(
                [stations, tstype],
                dtype=[("stations", stations.dtype), ("tstype", tstype.dtype)],
            )

            # set meta-ids for each variable but ensure that counter
            # is for all vars, var_metas contain (stations, tstype) tuples
            var_metas.append_var_station_tstype(var, station_tstype)
            if len(var_data) == 0:
                continue

            var_units[var] = var_data.units
            dra_data = {
                "meta_id": _station_tstype_to_int_array(station_tstype, var_metas[var]),
                "var_id": np.zeros(len(var_data), dtype="i2") + ugs.var_idx[var],
                "start_time": var_data.start_times,
                "end_time": var_data.end_times,
                "data": var_data.values,
                "stdev": var_data.standard_deviations,
                "dataaltitude": var_data.altitudes.astype("i2"),
                "flag": var_data.flags,  # TODO check for common undefined values?
            }
            ugs._dra.append_array(**dra_data)

        logger.info(f"Converting metadata from pyaro/{data_id} to ungridded")
        rev = None
        if "revision" in reader.metadata():
            rev = reader.metadata()["revision"]
        else:
            logger.warning(
                f"pyaro/{data_id} does not contain a 'revision', please inform data-provider, or pyaro-readers"
            )

        stations_with_metadata = reader.stations()
        for var in vars_to_retrieve:
            for station_tstype, meta_id in var_metas[var].items():
                (station_name, tstype) = station_tstype
                extra_metadata = stations_with_metadata[station_name].metadata
                d = {
                    "data_id": data_id,
                    "data_revision": rev,
                    "station_name": station_name,
                    "var_info": {
                        var: {"units": var_units[var]},
                    },
                    **stations_with_metadata[station_name],
                    **extra_metadata,
                }
                if "ts_type" not in d:
                    d["ts_type"] = tstype
                ugs.metadata[meta_id] = d

        logger.info(f"Finished converting from pyaro/{data_id} to UngriddedDataStructured")

        return ugs

    def clear_meta_no_data(self, inplace=True):
        """Remove all metadata blocks that do not have data associated with it

        Parameters
        ----------
        inplace : bool
            if True, the changes are applied to this instance directly, else
            to a copy

        Returns
        -------
        UngriddedData
            cleaned up data object

        Raises
        ------
        DataCoverageError
            if filtering results in empty data object
        """
        if inplace:
            obj = self
        else:
            obj = self.copy()

        meta_no_data = []
        # faster implementation of: distinct_metas = np.unique(obj._dra.data["meta_id"])
        distinct_metas = np.nonzero(np.bincount(obj._dra.data["meta_id"]))[0]

        for meta_idx, meta in obj.metadata.items():
            if not np.any(distinct_metas == meta_idx):
                # sanity check
                if bool(meta["var_info"]):
                    raise AttributeError(
                        "meta_idx {} suggests empty data block "
                        "but metadata[{}] contains variable "
                        "information"
                    )
                else:
                    meta_no_data.append(meta_idx)
        if len(meta_no_data):
            for meta_idx in meta_no_data:
                del obj.metadata[meta_idx]
        if len(obj.metadata) == 0:
            raise DataCoverageError("UngriddedData object appears to be empty")

        obj._add_to_filter_history(
            f"Removed {len(meta_no_data)} metadata blocks that have no data assigned"
        )
        return obj
